from concurrent.futures import Executor

from .futures import Future, wrap_if


class ExecutorWrapper(Executor):
    """
    A wrapping for Python executors to return Satella futures instead of standard Python ones.
    """

    def __init__(self, executor: Executor):
        self.executor = executor

    def submit(self, fn, *args, **kwargs) -> Future:
        return wrap_if(self.executor.submit(fn, *args, **kwargs))

    def shutdown(self, wait=True):
        """Clean-up the resources associated with the Executor.

        It is safe to call this method several times. Otherwise, no other
        methods can be called after this one.

        Args:
            wait: If True then shutdown will not return until all running
                futures have finished executing and the resources used by the
                executor have been reclaimed.
        """
        return self.executor.shutdown(wait=wait)



import unittest

from satella.coding.optionals import call_if_nnone, iterate_if_nnone, Optional, extract_optional


class TestOptionals(unittest.TestCase):

    def test_extract_optional(self):
        b = Optional(None)
        self.assertIs(extract_optional(b.test.value), None)
        self.assertEqual(extract_optional(5), 5)

    def test_optional(self):
        b = Optional(None)
        self.assertFalse(b.consumer.cancel().result())

    def test_iterate(self):
        b = list(iterate_if_nnone([1, 2, 3]))
        c = list(iterate_if_nnone(None))
        self.assertFalse(c)
        self.assertTrue(b)

    def test_optional_eq(self):
        class Opt:
            a = 5
            b = [2]

        a = Opt()
        self.assertEqual(Optional(a).a, 5)
        self.assertIn(2, Optional(a).b)
        self.assertNotEqual(Optional(None).a, 5)
        self.assertNotIn(2, Optional(None).b)

    def test_optional(self):
        self.assertIsNone(call_if_nnone(None))
        b = call_if_nnone(lambda y: y, 5)
        self.assertIsNotNone(b)
        self.assertEqual(5, b)



import typing as tp


class CountingDict:
    """
    A dictionary to quickly count the amount of elements in a set

    :param set_to_count: a sequence, whose elements should be counted.
        The elements should support being keys in a dictionary.
    """
    __slots__ = ('dct',)

    def __init__(self, set_to_count: tp.Sequence = ()):
        self.dct = {}
        for elem in set_to_count:
            self.count(elem)

    def __getitem__(self, item):
        return self.dct[item]

    def count(self, elem, pieces: int = 1) -> None:
        """
        Add an instance of elem to the set

        :param elem: instance to add
        :param pieces: amount to add
        """
        if elem not in self.dct:
            self.dct[elem] = pieces
        else:
            self.dct[elem] += pieces

    def clear(self):
        self.dct = {}

    def keys(self):
        return self.dct.keys()

    def values(self):
        return self.dct.values()

    def items(self):
        return self.dct.items()

    def __iter__(self):
        return iter(self.dct)

    def __len__(self):
        return len(self.dct)

    def __bool__(self):
        return bool(self.dct)



import typing as tp
from concurrent.futures.thread import ThreadPoolExecutor

from satella.coding.typing import V, U

try:
    import opentracing
except ImportError:
    opentracing = None


def parallel_construct(iterable: tp.Iterable[V],
                       function: tp.Callable[[V], tp.Optional[U]],
                       thread_pool: ThreadPoolExecutor,
                       span_title: tp.Optional[str] = None) -> tp.List[U]:
    """
    Construct a list from executing given function in a thread pool executor.

    If opentracing is installed, and tracing is enabled, current span will be passed to child threads.

    :param iterable: iterable to apply
    :param function: function to apply. If that function returns None, no element will be added
    :param thread_pool: thread pool to execute
    :param span_title: span title to create. For each execution a child span will be returned
    :return: list that is the result of parallel application of function on each element
    """
    wrap_iterable = None

    if opentracing is not None:
        tracer = opentracing.global_tracer()
        span = tracer.active_span
        if span is not None:

            def wrap_iterable(arg, *args,  **kwargs):
                with tracer.start_active_span(span_title or 'New span', child_of=span):
                    return function(arg, *args, **kwargs)

    if wrap_iterable is None:
        wrap_iterable = function

    result = []
    for item in thread_pool.map(wrap_iterable, iterable):
        if item is not None:
            result.append(item)
    return result



import inspect
import logging
import threading
import typing as tp

from .decorators.decorators import wraps
from .typing import ExceptionClassType, T, NoArgCallable, ExceptionList


def silence_excs(*exc_types: ExceptionClassType, returns=None,
                 returns_factory: tp.Optional[NoArgCallable[tp.Any]] = None):
    """
    Silence given exception types.

    Can be either a decorator or a context manager.

    If you are using it as a decorator, you can specify what value should the function return
    by using the returns kwarg:

    >>> @silence_excs(KeyError, returns=5)
    >>> def returns_5():
    >>>     raise KeyError()
    >>> assert returns_5() == 5

    Or if you want to you can specify a callable that will return the value you want to return

    >>> @silence_excs(KeyError, returns_factory=lambda: 5)
    >>> def returns_5():
    >>>     raise KeyError()
    >>> assert returns_5() == 5

    :raises ValueError: you gave both returns and returns_factory. You can only pass one of them!
    """
    return rethrow_as(exc_types, None, returns=returns,
                      returns_factory=returns_factory)


class log_exceptions:
    """
    Decorator/context manager to log your exceptions into the log.

    The exception will be logged and re-raised.

    Logger will be passed the exception instance as exc_info.

    :param logger: a logger to which the exception has to be logged
    :param severity: a severity level
    :param format_string: a format string with fields:
        - e      : the exception instance itself
        - args   : positional arguments with which the function was called, unavailable if context
                   manager
        - kwargs : keyword arguments with which the function was called, unavailable if context
                   manager
        You can specify additional fields providing the locals_ argument
        Example: "{exc_type} occurred with message {exc_val} with traceback {exc_tb}"
    :param locals_: local variables to add to the format string. args and kwargs will be overwritten
        by this, but e will never be overwritten.
    :param exc_types: logger will log only on those exceptions. Default is None which means
        log on all exceptions
    :param swallow_exception: if True, exception will be swallowed
    """
    __slots__ = ('logger', 'severity', 'format_string', 'locals', 'exc_types',
                 'swallow_exception')

    def __init__(self, logger: logging.Logger,
                 severity: int = logging.ERROR,
                 format_string: str = '{e}',
                 locals_: tp.Optional[tp.Dict] = None,
                 exc_types: tp.Union[
                     ExceptionClassType, tp.Sequence[ExceptionClassType]] = Exception,
                 swallow_exception: bool = False):
        self.logger = logger
        self.swallow_exception = swallow_exception
        self.severity = severity
        self.format_string = format_string
        self.locals = locals_
        self.exc_types = exc_types

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is not None:
            if not self.analyze_exception(exc_val, (), {}):
                return False
            else:
                return self.swallow_exception
        return False

    def analyze_exception(self, e, args, kwargs) -> bool:
        """Return whether the exception has been logged"""
        if isinstance(e, self.exc_types):
            format_dict = {'args': args,
                           'kwargs': kwargs}
            if self.locals is not None:
                format_dict.update(self.locals)
            format_dict['e'] = e
            self.logger.log(self.severity, self.format_string.format(**format_dict),
                            exc_info=e)
            return True
        return False

    def __call__(self, fun):
        if inspect.isgeneratorfunction(fun):
            @wraps(fun)
            def inner(*args, **kwargs):
                # noinspection PyBroadException
                try:
                    yield from fun(*args, **kwargs)
                except Exception as e:
                    if not self.analyze_exception(e, args, kwargs):
                        raise
                    elif not self.swallow_exception:
                        raise

            return inner
        else:
            @wraps(fun)
            def inner(*args, **kwargs):
                # noinspection PyBroadException
                try:
                    return fun(*args, **kwargs)
                except Exception as e:
                    if not self.analyze_exception(e, args, kwargs):
                        raise
                    elif not self.swallow_exception:
                        raise

            return inner


# noinspection PyPep8Naming
class reraise_as:
    """
    Transform some exceptions into others.

    Either a decorator or a context manager

    New exception will be created by calling exception to transform to with
    repr of current one.

    You can also provide just two exceptions, eg.

    >>> reraise_as(NameError, ValueError, 'a value error!')

    You can also provide a catch-all:

    >>> reraise_as((NameError, ValueError), OSError, 'an OS error!')

    New exception will be raised from the one caught!

    .. note:: This checks if exception matches directly via :code:`isinstance`, so defining
        your own subclassing hierarchy by :code:`__isinstance__` or :code:`__issubclass__`
        will work here.

    This is meant as an improvement of :class:`~satella.coding.rethrow_as`

    :param source_exc: source exception or a tuple of exceptions to catch
    :param target_exc: target exception to throw. If given a None, the exception will
        be silently swallowed.
    :param args: arguments to constructor of target exception
    :param kwargs: keyword arguments to constructor of target exception
    """
    __slots__ = 'source', 'target_exc', 'args', 'kwargs'

    def __init__(self, source_exc: ExceptionList,
                 target_exc: tp.Optional[ExceptionClassType], *args, **kwargs):
        self.source = source_exc
        self.target_exc = target_exc
        self.args = args
        self.kwargs = kwargs

    def __call__(self, fun: tp.Callable) -> tp.Callable:
        @wraps(fun)
        def inner(*args, **kwargs):
            try:
                return fun(*args, **kwargs)
            except Exception as e:
                if isinstance(e, self.source):
                    if self.target_exc is not None:
                        raise self.target_exc(*self.args, **self.kwargs) from e
                else:
                    raise

        return inner

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_val is not None:
            if isinstance(exc_val, self.source):
                if self.target_exc is None:
                    return True
                raise self.target_exc(*self.args, **self.kwargs) from exc_val
        return False


# noinspection PyPep8Naming
class rethrow_as:
    """
    Transform some exceptions into others.

    Either a decorator or a context manager

    New exception will be created by calling exception to transform to with
    repr of current one.

    .. note:: This checks if exception matches directly via :code:`isinstance`, so defining
        your own subclassing hierarchy by :code:`__isinstance__` or :code:`__issubclass__`
        will work here.

    You can also provide just two exceptions, eg.

    >>> rethrow_as(NameError, ValueError)

    You can also provide a pairwise translation, eg. from NameError to ValueError and from OSError
    to IOError

    >>> rethrow_as((NameError, ValueError), (OSError, IOError))

    If the second value is a None, exception will be silenced.

    Pass tuples of (exception to catch - exception to transform to).

    .. warning:: Try to use :class:`~satella.coding.reraise_as` instead.
        However, during to richer set of switches and capability to return a value
        this is not deprecated.

    :param exception_preprocessor: other callable/1 to use instead of repr.
        Should return a str, a text description of the exception
    :param returns: what value should the function return if this is used as a decorator
    :param returns_factory: a callable that returns the value this function should return is this
        is used as as decorator
    :raises ValueError: you specify both returns and returns_factory
    """
    __slots__ = 'mapping', 'exception_preprocessor', 'returns', '__exception_remapped', \
                'returns_factory'

    def __init__(self, *pairs: ExceptionList,
                 exception_preprocessor: tp.Optional[tp.Callable[[Exception], str]] = repr,
                 returns=None,
                 returns_factory: tp.Optional[NoArgCallable[tp.Any]] = None):
        try:
            a, b = pairs  # throws ValueError
            op = issubclass(b, BaseException)  # throws TypeError
        except TypeError:
            op = b is None
        except ValueError:
            op = False

        if op:
            pairs = [pairs]

        self.mapping = list(pairs)
        self.exception_preprocessor = exception_preprocessor or repr
        self.returns = returns
        self.returns_factory = returns_factory

        if self.returns is not None and self.returns_factory is not None:
            raise ValueError('You can specify only one of (returns, returns_factory)')

        # this is threading.local because two threads may execute the same function at the
        # same time, and exceptions from one function would leak to another
        self.__exception_remapped = threading.local()

    def __call__(self, fun: tp.Callable) -> tp.Callable:
        @wraps(fun)
        def inner(*args, **kwargs):
            with self:
                v = fun(*args, **kwargs)
            if self.__exception_remapped.was_raised:
                # This means that the normal flow of execution was interrupted
                if self.returns is not None:
                    return self.returns
                elif self.returns_factory is not None:
                    return self.returns_factory()
                return
            else:
                return v

        return inner

    def __enter__(self):
        self.__exception_remapped.was_raised = False
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is not None:
            for from_, to in self.mapping:
                if issubclass(exc_type, from_):
                    self.__exception_remapped.was_raised = True
                    if to is None:
                        return True
                    else:
                        raise to(self.exception_preprocessor(exc_val))


def raises_exception(exc_class: tp.Union[ExceptionClassType, tp.Tuple[ExceptionClassType, ...]],
                     clb: NoArgCallable[None]) -> bool:
    """
    Does the callable raise a given exception?
    """
    try:
        clb()
    except exc_class:
        return True
    else:
        return False


def catch_exception(exc_class: tp.Union[ExceptionClassType, tp.Tuple[ExceptionClassType, ...]],
                    clb: NoArgCallable[tp.Optional[T]],
                    return_instead: tp.Optional[T] = None,
                    return_value_on_no_exception: bool = False) -> tp.Union[Exception, T]:
    """
    Catch exception of given type and return it. Functionally equivalent to:

    >>> try:
    >>>     v = clb()
    >>>     if return_value_on_no_exception:
    >>>         return v
    >>> except exc_class as e:
    >>>     if return_instead:
    >>>         return return_instead
    >>>     return e

    If a different class of exception is caught, it will be propagated.

    :param exc_class: Exception classes to catch
    :param clb: callable/0 to call to raise the exception
    :param return_instead: what to return instead of the function result if it didn't end in an
        exception
    :param return_value_on_no_exception: whether to return the function result if exception didn't
        happen
    :raises ValueError: an exception was not thrown
    """
    try:
        result = clb()
    except exc_class as e:
        return e

    if return_instead is not None:
        return return_instead

    if return_value_on_no_exception:
        return result

    raise ValueError('Callable executed without error')



from .enums import ComparableEnum, ComparableIntEnum, HashableIntEnum
from .hashable import ComparableAndHashableBy, ComparableAndHashableByInt, \
    OmniHashableMixin, HashableMixin, ComparableAndHashableByStr
from .strings import ReprableMixin, StrEqHashableMixin
from .eqable import DictionaryEQAble

__all__ = ['ComparableIntEnum', 'ComparableEnum', 'ComparableAndHashableBy',
           'HashableIntEnum', 'ComparableAndHashableByInt', 'OmniHashableMixin',
           'ReprableMixin', 'StrEqHashableMixin', 'HashableMixin',
           'ComparableAndHashableByStr', 'DictionaryEQAble']



from .exceptions import trace_exception, set_maximum_traceback_length
from .trace import trace_future, trace_function

__all__ = ['trace_future', 'trace_function', 'trace_exception', 'set_maximum_traceback_length']



import gc
import signal
import sys
import typing as tp


def dump_memory_on(output: tp.TextIO = sys.stderr):
    """
    Dump statistics about current Python memory usage to target stream.

    Each Python object will be printed, along with a breakdown of most types and their total usage.

    Make sure you have enough memory to generate a breakdown. You can preallocate something at the
    start for example.

    .. warning:: This will return size of 0 on PyPy

    :param output: output, default is stderr
    """
    top_scores = {}
    instances = {}
    for obj in gc.get_objects():
        typ = type(obj)
        try:
            size = sys.getsizeof(obj)
        except TypeError:
            size = 0
        if typ in top_scores:
            top_scores[typ] += size
            instances[typ] += 1
        else:
            top_scores[typ] = size
            instances[typ] = 1
        output.write('object %s type %s size %s bytes\n' % (repr(obj), typ, size))

    output.write('----------------------------------\n')
    output.write('Memory usage scores: \n')
    output.write('----------------------------------\n')
    items = list(top_scores.items())
    items.sort(key=lambda y: -y[1])
    for typ, tot_size in items:
        output.write('%s: %s bytes %s instances\n' % (typ, tot_size, instances[typ]))


def install_dump_memory_on(signal_number, output: tp.TextIO = sys.stderr):
    """
    Instruct Python to dump all frames onto output, along with their local variables
    upon receiving given signal

    :param signal_number: number of the signal
    :param output: output
    """
    signal.signal(signal_number,
                  lambda sig_no, stack_frame: dump_memory_on(output))



from satella.configuration import sources
from satella.configuration.sources import EnvVarsSource, OptionalSource, AlternativeSource, \
    EnvironmentSource, \
    StaticSource, MergingSource
from .utils import SourceTestCase, mock_env


class TestEnvVarsSource(SourceTestCase):
    @mock_env('satella', '{"a":2}')
    def test_ok(self):
        self.assertSourceHas(sources.EnvVarsSource('satella'), {u"a": 2})

    def test_none(self):
        self.assertSourceEmpty(OptionalSource(EnvVarsSource('satella')))
        self.assertSourceFails(EnvVarsSource('satella'))

    def test_alternative(self):
        ds = AlternativeSource(
            EnvironmentSource('test', 'test'),
            StaticSource({'test': 'test'})
        )
        self.assertSourceHas(ds, {'test': 'test'})

    def test_merging(self):
        ds = MergingSource(
            EnvironmentSource('test', 'test'),
            on_fail=MergingSource.SILENT,
            fail_if_no_sources_are_correct=False
        )
        self.assertEqual(ds.provide(), {})



import itertools
import typing as tp
import warnings

from satella.coding.decorators.decorators import wraps
from satella.coding.typing import ExceptionClassType


def retry(times: tp.Optional[int] = None,
          exc_classes: tp.Union[ExceptionClassType, tp.Tuple[ExceptionClassType, ...]] = Exception,
          on_failure: tp.Callable[[Exception], None] = lambda e: None,
          swallow_exception: bool = True,
          call_on_failure: tp.Optional[tp.Callable[[Exception], None]] = None,
          call_on_success: tp.Optional[tp.Callable[[int], None]] = None,
          call_between_retries: tp.Optional[tp.Callable[[Exception], None]] = None):
    """
    A decorator retrying given operation, failing it when an exception shows up.

    Essentially this:

    >>> @retry(3, RuntimeError)
    >>> def do_op():
    >>>     .. do op ..

    Is a syntactic sugar for this:

    >>> for i in range(3):
    >>>     try:
    >>>         .. do op ..
    >>>         break
    >>>     except RuntimeError:
    >>>         pass

    .. warning:: Retry by default will swallow the resulting exception. To avoid
        this behaviour set the :code:`swallow_exception` parameter to False.

    :param times: maximum times to retry this operation. By default (None) retry until the world ends.
    :param exc_classes: exception classes that should be considered failure. By default.
        catches everything (Exception).
    :param on_failure: callable to call when it fails times times. Callable will be called
        with a single argument, exception instance that was raised last. That exception will
        be swallowed, unless swallow_exception is set to False
    :param swallow_exception: the last exception will be swallowed, unless this is set to False
    :param call_on_failure: a callable that will be called upon failing to do this, with an
        exception as it's sole argument. It's result will be discarded. Deprecated, use on_failure
        instead.
    :param call_on_success: a callable that will be called with a single argument: the number
        of retries that it took to finish the job. It's result will be discarded.
    :param call_between_retries: called between retries with a single argument, the Exception
        instance that forced the retry.
    :return: function result
    """

    if call_on_failure is not None:
        warnings.warn('call_on_failure is deprecated, use on_failure instead', DeprecationWarning)

    def outer(fun):
        @wraps(fun)
        def inner(*args, **kwargs):
            if times is None:
                iterator = itertools.count()
            else:
                iterator = range(times)
            for i in iterator:
                try:
                    y = fun(*args, **kwargs)
                    if call_on_success is not None:
                        call_on_success(i)
                    return y
                except exc_classes as e:
                    f = e
                    if call_between_retries is not None:
                        call_between_retries(e)
                    continue
            else:
                on_failure(f)
                if call_on_failure is not None:
                    call_on_failure(f)
                if not swallow_exception:
                    raise f

        return inner

    return outer



import queue
import unittest

from satella.coding.structures import Singleton, SingletonWithRegardsTo, \
    get_instances_for_singleton, delete_singleton_for
from satella.coding import queue_iterator


class TestSingleton(unittest.TestCase):

    def test_queue_iterator(self):
        q = queue.Queue()
        q.put(2)
        q.put(5)
        self.assertEqual(list(queue_iterator(q)), [2, 5])

    def test_singleton(self):
        @Singleton
        class MyClass(object):
            def __init__(self):
                self.a = 5

        a = MyClass()
        b = MyClass()

        a.a = 6

        self.assertEqual(b.a, 6)

    def test_singleton_taking_arguments(self):
        @Singleton
        class MyClass(object):
            def __init__(self, a):
                self.a = a

        a = MyClass(5)
        b = MyClass(6)

        self.assertEqual(a.a, 5)
        self.assertEqual(b.a, 5)

    def test_singleton_with_regards_to(self):
        @SingletonWithRegardsTo(num_args=1)
        class MyClass:
            def __init__(self, device_id: str):
                self.device_id = device_id

        a = MyClass('a')
        b = MyClass('b')
        c = MyClass('a')
        self.assertIs(a, c)
        self.assertIsNot(a, b)
        self.assertEqual(a.device_id, 'a')
        self.assertEqual(b.device_id, 'b')
        a.device_id = 'c'
        self.assertEqual(c.device_id, 'c')
        self.assertEqual(set(get_instances_for_singleton(MyClass)), {('a',), ('b',)})
        delete_singleton_for(MyClass, 'a')
        self.assertEqual(set(get_instances_for_singleton(MyClass)), {('b',)})



from __future__ import absolute_import
import struct
import typing as tp

from satella.exceptions import NotEnoughBytes


class BinaryParser:
    """
    A class that allows parsing binary streams easily.

    This supports __len__ to return the amount of bytes in the stream,
    and __bytes__ to return the bytes.

    This is a zero-copy solution, and :meth:`get_parser` will be zero copy
    as well.

    :param b_stream: an object that allows subscripts to
        span ranges, which will return items parsable by struct
    :param offset: initial offset into the stream
    :param length: optional maximum length of byte count
    :raises NotEnoughBytes: offset larger than stream length

    :ivar pointer: pointer to the next bytes. Can be read and modified at will to
        preserve the earlier state of the BinaryParser.
    """
    def __len__(self) -> int:
        return self.length

    def get_remaining_bytes_count(self) -> int:
        """
        Return the amount of bytes remaining. This will not advance the pointer
        """
        return self.length - self.pointer + self.init_ofs

    def __init__(self, b_stream: tp.Union[bytes, bytearray], offset: int = 0,
                 length: tp.Optional[int] = None):
        self.b_stream = b_stream
        self.pointer = self.init_ofs = offset
        self.struct_cache = {}
        self.length = length or len(b_stream)
        if offset > len(self.b_stream):
            raise NotEnoughBytes('Offset larger than the stream!')

    def __bytes__(self) -> bytes:
        return self.b_stream[self.init_ofs:self.init_ofs+self.length]

    def skip(self, n: int) -> None:
        """
        Advance the pointer by n bytes

        :param n: bytes to advance
        :raises NotEnoughBytes: not enough bytes remain in the stream!
        """
        self.assert_has_bytes(n)
        self.pointer += n

    def assert_has_bytes(self, n: int) -> None:
        """
        Assert that we have at least n bytes to consume.

        This does not advance the pointer.

        :param n: amount of bytes to consume
        :raises NotEnoughBytes: not enough bytes remain in the stream!
        """
        if self.length + self.init_ofs < self.pointer + n:
            raise NotEnoughBytes('Not enough bytes')

    def get_parser(self, length: int) -> 'BinaryParser':
        """
        Return a subclassed binary parser providing a window to another binary parser's data.

        This will advance the pointer by length bytes

        :param length: amount of bytes to view
        :return: a BinaryParser
        :raises NotEnoughBytes: not enough bytes remain in the stream!
        """
        self.assert_has_bytes(length)
        try:
            return BinaryParser(self.b_stream, self.pointer, length)
        finally:
            self.pointer += length

    def reset(self) -> None:
        """
        Reset the internal pointer to starting value
        :return:
        """
        self.pointer = self.init_ofs

    def _to_struct(self, st: tp.Union[str, struct.Struct]) -> struct.Struct:
        if isinstance(st, struct.Struct):
            return st
        else:
            if st in self.struct_cache:
                fmt = st
                return self.struct_cache[fmt]
            else:
                fmt = st
                st = struct.Struct(st)
                self.struct_cache[fmt] = st
                return st

    def get_bytes(self, n: int) -> bytes:
        """
        Return this many bytes

        :param n: amount of bytes to return
        :return: bytes returned
        :raises NotEnoughBytes: not enough bytes remain in the stream!
        """
        self.assert_has_bytes(n)
        try:
            return self.b_stream[self.pointer:self.pointer+n]
        finally:
            self.pointer += n

    def get_struct(self, st: tp.Union[str, struct.Struct]) -> tp.Union[int, float]:
        """
        Try to obtain as many bytes as this struct requires and return them parsed.

        This must be a single-character struct!

        This will advance the pointer by size of st. Struct objects
        will be served from internal instance-specific cache.

        :param st: a single-character struct.Struct or a single character struct specification
        :return: a value returned from it
        :raises NotEnoughBytes: not enough bytes remain in the stream!
        :raises AssertionError: struct was not a single element one!
        """
        st = self._to_struct(st)

        if st.format[0] in {'>', '<', '!', '@'}:
            assert len(st.format) == 2, 'Format must span at most 1 character, use ' \
                                        'get_structs for multiples!'
        else:
            assert len(st.format) == 1, 'Format must span at most 1 character, use ' \
                                        'get_structs for multiples!'

        st_len = st.size
        self.assert_has_bytes(st_len)
        try:
            return st.unpack(self.b_stream[self.pointer:self.pointer+st_len])[0]
        finally:
            self.pointer += st_len

    def get_structs(self, st: tp.Union[str, struct.Struct]) -> tp.Tuple[tp.Union[int, float],
                                                                        ...]:
        """
        Try to obtain as many bytes as this struct requires and return them parsed.

        This will advance the pointer by size of st. Struct objects
        will be served from internal instance-specific cache.

        :param st: a struct.Struct or a multi character struct specification
        :return: a tuple of un-parsed values
        :raises NotEnoughBytes: not enough bytes remain in the stream!
        """
        st = self._to_struct(st)
        st_len = st.size
        self.assert_has_bytes(st_len)
        try:
            return st.unpack(self.b_stream[self.pointer:self.pointer+st_len])
        finally:
            self.pointer += st_len

    def get_remaining_bytes(self) -> tp.Union[bytes, bytearray]:
        """
        Return the remaining bytes.

        This will not advance the pointer
        """
        advance = self.pointer - self.init_ofs
        remaining = self.length - advance
        return self.b_stream[self.pointer:self.pointer+remaining]



import os
import signal
import sys
import unittest
import time
import multiprocessing

from satella.os import hang_until_sig


class TestHangUntilSig(unittest.TestCase):

    @unittest.skipIf('win' in sys.platform, 'Needs a POSIX to run')
    def test_hang_until_sig(self):

        def child_process():
            time.sleep(1)
            os.kill(os.getppid(), signal.SIGTERM)

        mp = multiprocessing.Process(target=child_process)
        mp.start()
        try:
            hang_until_sig()
        finally:
            os.wait()



from ..immutable import NOT_EQUAL_TO_ANYTHING


class DictionaryEQAble:
    """
    A class mix-in that defines __eq__ and __ne__ to be:

    - both the same exact type (so subclassing won't work)
    - have the exact same __dict__
    """

    def __eq__(self, other) -> bool:
        if type(self) != type(other):
            return False
        for key in self.__dict__.keys():
            if getattr(other, key, NOT_EQUAL_TO_ANYTHING) != self.__dict__[key]:
                return False
        return True

    def __ne__(self, other) -> bool:
        if type(self) != type(other):
            return True
        for key in self.__dict__.keys():
            if getattr(other, key, NOT_EQUAL_TO_ANYTHING) != self.__dict__[key]:
                return True
        return True



import os
import tempfile
import unittest

from satella.coding import silence_excs
from satella.configuration.sources import FileSource, FORMAT_SOURCES, \
    DirectorySource
from .utils import SourceTestCase


class TestFileSource(SourceTestCase):
    def setUp(self):
        super(TestFileSource, self).setUp()
        self.filename = tempfile.mktemp()

    @silence_excs(FileNotFoundError)
    def tearDown(self):
        os.unlink(self.filename)

    def check_with_data(self, data: bytes, value: dict):
        with open(self.filename, 'wb') as fout:
            fout.write(data)

        self.assertSourceHas(FileSource(self.filename), value)

    def test_json(self):
        self.check_with_data(b'{"a": 2}', {"a": 2})

    @unittest.skipIf('YAMLSource' not in FORMAT_SOURCES,
                     reason='No pyyaml installed')
    def test_yaml(self):
        self.check_with_data(b"\na: 5\n", {"a": 5})

    @unittest.skipIf('TOMLSource' not in FORMAT_SOURCES,
                     reason='No TOML installed')
    def test_toml(self):
        self.check_with_data(b"a = 5", {"a": 5})


class TestDirectorySource(SourceTestCase):
    def test_directory_invalids(self):
        with tempfile.TemporaryDirectory() as outdir:
            self.filename = outdir
            self.makefile("valid-json.json", b'''{
  "amqp": "amqp",
  "logstash": {
    "host": "localhost",
    "port": 9600
  }
}
''')
            self.makefile("invalid-json.json", b'''{"CI": true''')
            self.ds = DirectorySource(outdir, on_fail=DirectorySource.SILENT)
            self.assertEqual(self.ds.provide(),
                             {'amqp': 'amqp', 'logstash': {'host': 'localhost', 'port': 9600}})

    def test_directory_source(self):
        with tempfile.TemporaryDirectory() as outdir:
            self.filename = outdir
            self.assertTrue(os.path.isdir(outdir))
            self.ds = DirectorySource(outdir, scan_subdirectories=True)
            self.makefile('config.json', b'{"a": [2]}')
            self.makefile('config2.json', b'{"a": [2]}')
            os.mkdir(os.path.join(outdir, 'temp1'))
            self.makefile(os.path.join(outdir, 'temp1', 'config3.json'),
                          b'{"a": [23]}')
            self.assertSourceHas(self.ds, set((2, 2, 23)), postop=lambda x: set(x['a']))

    def makefile(self, name: str, value: bytes):
        with open(os.path.join(self.filename, name), 'wb') as fout:
            fout.write(value)



import abc
import collections
import copy
import math
import time
import unittest

from unittest import mock

from satella.coding.concurrent import call_in_separate_thread
from satella.coding.structures import TimeBasedHeap, Heap, typednamedtuple, \
    OmniHashableMixin, DictObject, apply_dict_object, Immutable, frozendict, SetHeap, \
    DictionaryView, HashableWrapper, TwoWayDictionary, Ranking, SortedList, SliceableDeque, \
    DirtyDict, KeyAwareDefaultDict, Proxy, ReprableMixin, TimeBasedSetHeap, ExpiringEntryDict, SelfCleaningDefaultDict, \
    CacheDict, StrEqHashableMixin, ComparableIntEnum, HashableIntEnum, ComparableAndHashableBy, \
    ComparableAndHashableByInt, SparseMatrix, ExclusiveWritebackCache, Subqueue, \
    CountingDict, ComparableEnum, LRU, LRUCacheDict, Vector, DefaultDict, PushIterable, \
    ComparableAndHashableByStr, NotEqualToAnything, NOT_EQUAL_TO_ANYTHING, DictionaryEQAble


class TestStructures(unittest.TestCase):

    def test_dictionary_eqable(self):
        class Dupa(DictionaryEQAble):
            def __init__(self, a):
                self.a = a

        class Dupa2(DictionaryEQAble):
            def __init__(self, a):
                self.a = a

        a = Dupa(5)
        b = Dupa(5)
        c = Dupa(6)
        d = Dupa2(5)
        self.assertTrue(a == b)
        self.assertFalse(a == c)
        self.assertFalse(a == d)

    def test_not_equal_to_anything(self):
        self.assertTrue(NOT_EQUAL_TO_ANYTHING != NOT_EQUAL_TO_ANYTHING)
        self.assertFalse(NOT_EQUAL_TO_ANYTHING == NOT_EQUAL_TO_ANYTHING)

        class Dupa(NotEqualToAnything):
            pass

        a = Dupa()

        self.assertTrue(a != a)
        self.assertFalse(a == a)

    def test_push_iterable(self):

        a = PushIterable([1, 2, 3])
        self.assertEqual(a.pop(), 1)
        a.push(0)
        self.assertEqual(a.pop(), 0)
        self.assertEqual(a.pop(), 2)
        a.push(0)
        a.push(1)
        self.assertEqual(a.pop(), 1)
        self.assertEqual(a.pop(), 0)
        a.push_left(0)
        a.push_left(1)
        self.assertEqual(a.pop(), 0)
        self.assertEqual(a.pop(), 1)
        self.assertEqual(a.pop(), 3)
        self.assertRaises(StopIteration, a.pop)

    def test_default_dict(self):
        a = DefaultDict(lambda: '')
        a[2]
        self.assertNotIn(2, a)

    def test_vector(self):
        a = Vector((2, 3))
        self.assertEqual(a-Vector((1, 2)), Vector((1, 1)))
        self.assertEqual(a*2, Vector((4, 6)))
        self.assertEqual(a/2, Vector((1, 1.5)))

    def test_lru_cache_dict(self):
        class TestCacheGetter:
            def __init__(self):
                self.called_on = {}
                self.value = 2

            def __call__(self, key):
                if self.value is None:
                    raise KeyError('no value available')
                self.called_on[key] = time.monotonic()
                return self.value

        cg = TestCacheGetter()
        cd = LRUCacheDict(1, 2, cg, max_size=3)

        self.assertEqual(cd[1], 2)
        self.assertEqual(cd[2], 2)
        self.assertEqual(cd[3], 2)
        self.assertEqual(cd[4], 2)
        self.assertEqual(len(cd), 3)

    def test_lru(self):
        lru = LRU()
        lru.add('a')
        lru.add('b')
        lru.add('c')
        self.assertEqual(lru.get_item_to_evict(), 'a')
        lru.mark_as_used('b')
        self.assertEqual(lru.get_item_to_evict(), 'c')
        self.assertEqual(len(lru), 1)

    def test_comparable_enum(self):
        class MyEnum(ComparableEnum):
            A = 'test'
            B = 'test2'

        self.assertEqual(MyEnum.A, 'test')
        self.assertNotEqual(MyEnum.B, 'test')

    def test_counting_dict(self):
        cd = CountingDict([1, 1, 2, 2, 3])
        self.assertEqual(cd[1], 2)
        self.assertEqual(list(cd), [1, 2, 3])
        cd.clear()
        self.assertFalse(cd)

    def test_subqueue(self):
        a = Subqueue()
        a.put('test', 2)
        self.assertEqual(a.get('test'), 2)

        @call_in_separate_thread()
        def put_a_message():
            nonlocal a
            time.sleep(1)
            a.put('test', 5)

        put_a_message()
        self.assertEqual(a.get('test'), 5)
        a.put('test2', 4)
        self.assertEqual(a.qsize(), 1)
        self.assertEqual(a.get_any(), ('test2', 4))

    def test_exclusive_writeback_cache(self):
        a = {5: 3, 4: 2, 1: 0}
        b = {'no_calls': 0}

        def setitem(k, v):
            nonlocal a, b
            b['no_calls'] += 1
            a[k] = v

        def getitem(k):
            nonlocal a, b
            b['no_calls'] += 1
            return a[k]

        def delitem(k):
            nonlocal a, b
            b['no_calls'] += 1
            del a[k]

        wbc = ExclusiveWritebackCache(setitem, getitem, delitem)
        self.assertEqual(wbc[5], 3)
        self.assertEqual(b['no_calls'], 1)
        self.assertRaises(KeyError, lambda: wbc[-1])
        self.assertEqual(b['no_calls'], 2)
        self.assertEqual(wbc[5], 3)
        self.assertEqual(b['no_calls'], 2)
        wbc[5] = 2
        wbc.sync()
        self.assertEqual(a[5], 2)
        self.assertEqual(b['no_calls'], 3)
        del wbc[4]
        wbc.sync()
        self.assertRaises(KeyError, lambda: a[4])

    def test_sparse_matrix(self):
        sm = SparseMatrix()
        sm[1, 2] = 1
        sm[0, 0] = 2
        self.assertEqual(list(sm), [[2, None], [None, None], [None, 1]])
        self.assertEqual(sm.columns, 2)
        self.assertEqual(sm.rows, 3)
        sm[1, 2] = 3
        self.assertEqual(list(sm), [[2, None], [None, None], [None, 3]])
        del sm[1, 2]
        a = list(sm)
        self.assertEqual(a, [[2]])
        sm[:, 2] = [5, 6, 7]
        self.assertEqual(list(sm), [[2, None, None], [None, None, None], [5, 6, 7]])
        sm[1, :] = [1, 2, 3]
        self.assertEqual(sm[1, :], [1, 2, 3])
        self.assertEqual(sm[:, 1], [None, 2, None])
        self.assertEqual(list(sm), [[2, 1, None], [None, 2, None], [5, 3, 7]])
        del sm[1, 2]
        self.assertEqual(list(sm), [[2, 1, None], [None, 2, None], [5, None, 7]])
        self.assertEqual(sm[-1, -1], 7)
        sm2 = SparseMatrix.from_iterable([[2, 1, None], [None, 2, None], [5, None, 7]])
        self.assertEqual(sm, sm2)
        del sm[1, :]
        self.assertEqual(list(sm), [[2, None, None], [None, None, None], [5, None, 7]])
        del sm[:, 2]
        self.assertEqual(list(sm), [[2]])
        del sm[:, :]
        self.assertEqual(list(sm), [])
        sm.append_row([1, 2])
        self.assertEqual(list(sm), [[1, 2]])
        sm = SparseMatrix([[1, 2], [3, 4]])
        self.assertEqual(list(sm), [[1, 2], [3, 4]])
        sm2 = sm.shoot()
        self.assertEqual(list(sm2), [[1, None, 2], [None, None, None], [3, None, 4]])
        self.assertEqual(set(sm2.get_neighbour_coordinates(0, 0)), {(0, 1), (1, 0), (1, 1)})
        self.assertEqual(set(sm2.get_neighbour_coordinates(1, 1, False)), {(0, 1), (1, 0), (2, 1), (1, 2)})
        self.assertEqual(set(sm2.get_neighbour_coordinates(2, 2)), {(2, 1), (1, 1), (1, 2)})
        self.assertEqual(sm2.min(), 1)
        self.assertEqual(sm2.max(), 4)
        self.assertRaises(IndexError, lambda: sm2[3, 3])

    def test_comparable_and_hashable_by_int(self):
        class MyClass(ComparableAndHashableByInt):
            def __init__(self, a: int):
                self.a = a

            def __int__(self):
                return self.a

        a = MyClass(1)
        b = MyClass(1)
        c = MyClass(2)
        self.assertEqual(a, b)
        self.assertNotEqual(a, c)
        self.assertLess(a, c)
        self.assertGreaterEqual(a, b)
        self.assertGreaterEqual(c, b)
        self.assertEqual(hash(a), hash(b))

    def test_comparable_and_hashable_by(self):
        class Vector(ComparableAndHashableBy):
            _COMPARABLE_BY = 'a'

            def __init__(self, a):
                self.a = a

        self.assertTrue(Vector(1) < Vector(2))
        self.assertTrue(Vector(1) <= Vector(2))
        self.assertFalse(Vector(1) > Vector(2))
        self.assertFalse(Vector(1) >= Vector(2))
        self.assertFalse(Vector(1) == Vector(2))
        self.assertEqual(hash(Vector(1)), hash(1))

    def test_comparable_and_hashable_by_str(self):
        class Vector(ComparableAndHashableByStr):

            def __str__(self) -> str:
                return self.a

            def __init__(self, a):
                self.a = a

        self.assertTrue(Vector('a') < Vector('ab'))
        self.assertTrue(Vector('a') <= Vector('a'))
        self.assertFalse(Vector('a') > Vector('b'))
        self.assertFalse(Vector('a') >= Vector('b'))
        self.assertFalse(Vector('a') == Vector('b'))
        self.assertEqual(hash(Vector('a')), hash('a'))

    def test_hashable_int_enum(self):
        class A(HashableIntEnum):
            A = 0
            B = 1

        self.assertEqual({A.A: 't', A.B: 'c'}[A.A], 't')

    def test_comparable_enum(self):
        class A(ComparableEnum):
            A = 1

        class B(ComparableEnum):
            A = 1

        self.assertEqual(A.A, B.A)

    def test_comparable_int_enum_bug1(self):
        def test_comparable_int_enum(self):
            A = 0
            B = 1
            C = 2
            class Enum(ComparableIntEnum):
                A = A
                B = B
                C = C

            self.assertTrue(Enum.A < B)
            self.assertTrue(Enum.A <= B)
            self.assertTrue(Enum.A <= A)
            self.assertFalse(Enum.A > B)
            self.assertFalse(Enum.A >= B)
            self.assertFalse(Enum.A == B)
            self.assertEqual({Enum.A: 't'}[A], 't')

    def test_comparable_int_enum(self):

        class Enum(ComparableIntEnum):
            A = 0
            B = 1
            C = 2

        self.assertTrue(Enum.A < Enum.B)
        self.assertTrue(Enum.A <= Enum.B)
        self.assertTrue(Enum.A <= Enum.A)
        self.assertFalse(Enum.A > Enum.B)
        self.assertFalse(Enum.A >= Enum.B)
        self.assertFalse(Enum.A == Enum.B)
        self.assertEqual({Enum.A: 't'}[Enum.A], 't')

    def test_str_eq_hashable_mixin(self):
        class MyClass(StrEqHashableMixin):
            def __init__(self, v: str):
                self.v = v

            def __str__(self) -> str:
                return self.v

        a = MyClass('a')
        self.assertEqual(hash('a'), hash(a))
        self.assertEqual(a, 'a')

    def test_proxy(self):
        a = Proxy(2)
        self.assertEqual(int(a), 2)
        self.assertEqual(a, 2)
        self.assertEqual(a+2, 4)
        self.assertEqual(a*2, 4)
        self.assertEqual(a-2, 0)
        self.assertIsNotNone(hash(a))
        self.assertIsNotNone(repr(a))
        self.assertEqual(str(a), '2')
        self.assertEqual(abs(a), 2)
        self.assertEqual(a/2, 1)
        self.assertEqual(a//2, 1)
        a += 1
        self.assertEqual(a, 3)
        a -= 1
        self.assertEqual(a, 2)
        a *= 2
        self.assertEqual(a, 4)
        a /= 2
        self.assertEqual(a, 1.0)
        a = Proxy(2)
        a //= 2
        self.assertEqual(a, 1)
        a <<= 1
        self.assertEqual(a, 2)
        a >>= 1
        self.assertEqual(a, 1)
        a ^= 1
        self.assertEqual(a, 0)
        self.assertEqual(a ^ 2, 2)
        a = Proxy(2)
        self.assertEqual(a << 1, 4)
        self.assertEqual(a >> 1, 1)
        a = Proxy(2)
        self.assertEqual(a ** 2, 4)
        a **= 2
        self.assertEqual(a, 4)
        self.assertEqual(-a, 4)
        self.assertEqual(+a, 4)
        self.assertTrue(a > 1)
        self.assertFalse(a < 1)
        self.assertTrue(a >= 1)
        self.assertFalse(a <= 1)

        b = Proxy(list)
        self.assertEqual(b((1, 2, 3)), [1, 2, 3])
        self.assertIsNotNone(b.insert)

        c = Proxy([1, 2, 3])
        self.assertEqual(len(c), 3)
        self.assertEqual(list(iter(c)), [1, 2, 3])
        self.assertEqual(c[0], 1)
        del c[0]
        self.assertEqual(c, [2, 3])
        c.append(4)
        self.assertEqual(c, [2, 3, 4])
        self.assertIn(2, c)

    def test_cache_dict(self):
        class TestCacheGetter:
            def __init__(self):
                self.called_on = {}
                self.value = 2

            def __call__(self, key):
                if self.value is None:
                    raise KeyError('no value available')
                self.called_on[key] = time.monotonic()
                return self.value

        cg = TestCacheGetter()
        cd = CacheDict(1, 2, cg)
        now = time.monotonic()
        self.assertEqual(cd[2], 2)
        self.assertEqual(list(cd), [2])
        self.assertGreaterEqual(cg.called_on[2], now)
        time.sleep(0.2)
        now = time.monotonic()
        self.assertLessEqual(cg.called_on[2], now)
        time.sleep(0.9)
        now = time.monotonic()
        self.assertEqual(cd[2], 2)
        time.sleep(0.2)
        self.assertGreaterEqual(cg.called_on[2], now)
        self.assertEqual(cd[2], 2)
        time.sleep(2.9)
        cg.value = 3
        now = time.monotonic()
        self.assertEqual(cd[2], 3)
        self.assertGreaterEqual(cg.called_on[2], now)
        del cd[2]
        self.assertEqual(len(cd), 0)
        cg.value = None
        self.assertRaises(KeyError, lambda: cd[2])
        cg.value = 2
        self.assertEqual(cd[2], 2)
        cg.value = None
        time.sleep(1.5)
        self.assertEqual(cd[2], 2)
        time.sleep(0.6)
        self.assertRaises(KeyError, lambda: cd[2])

        cd = CacheDict(1, 2, cg, cache_failures_interval=1)
        self.assertRaises(KeyError, lambda: cd[2])
        cg.value = 2
        self.assertRaises(KeyError, lambda: cd[2])
        time.sleep(1.1)
        self.assertEqual(cd[2], 2)
        cd.feed(5, 6)
        self.assertEqual(cd[5], 6)

    def test_cache_dict_default_value_factory(self):
        class TestCacheGetter:
            def __call__(self, key):
                raise KeyError('no value available')

        cg = TestCacheGetter()
        cd = CacheDict(1, 2, cg, cache_failures_interval=1, default_value_factory=list)
        self.assertEqual(cd[2], [])
        self.assertEqual(cd[2], [])

    def test_dictobject_dictobject(self):
        a = DictObject(a=5, k=3)
        b = DictObject(a)
        b.c = 4
        self.assertEqual(b.a, 5)
        self.assertEqual(b.k, 3)
        self.assertEqual(b.c, 4)

        self.assertIn('DictObject', str(b))
        self.assertIn('DictObject', repr(b))
        self.assertIn('a', b.keys())
        self.assertIn(5, b.values())
        self.assertIn(('a', 5), b.items())

    def test_dictobject_setdefault(self):
        a = DictObject()
        self.assertEqual(a.setdefault('k', 2), 2)
        self.assertEqual(a.k, 2)

        self.assertIsNone(a.get('v'))
        self.assertEqual(a.get('k'), 2)
        self.assertIn('k', a)

    def test_expiration_dict_manual_expiring(self):
        eed = ExpiringEntryDict(expiration_timeout=5)
        eed['test'] = 2
        time.sleep(2)
        self.assertEqual(eed['test'], 2)
        time.sleep(4)
        self.assertRaises(KeyError, lambda: eed['test'])

    def test_expiration_dict_self_expiring(self):
        eed = ExpiringEntryDict(expiration_timeout=5, external_cleanup=True)
        eed['test'] = 2
        time.sleep(2)
        self.assertEqual(eed['test'], 2)
        time.sleep(10)
        self.assertRaises(KeyError, lambda: eed.data['test'])

    def test_self_cleaning_default_dict_no_background_maintenance(self):
        sc_dd = SelfCleaningDefaultDict(list, False)
        sc_dd['test'].append(2)
        sc_dd['test'].pop()
        sc_dd['test'] = '2'
        self.assertEqual(len(sc_dd), 1)
        del sc_dd['test']
        self.assertEqual(len(sc_dd), 0)

    def test_self_cleaning_default_dict_background_maintenance(self):
        sc_dd = SelfCleaningDefaultDict(list)
        sc_dd['test'] = [1]
        sc_dd['test'].pop()
        time.sleep(10)
        self.assertEqual(len(sc_dd), 0)
        sc_dd['test'] = '2'
        self.assertEqual(len(sc_dd), 1)
        del sc_dd['test']
        self.assertEqual(len(sc_dd), 0)

    def test_reprable_mixin(self):
        class Test(ReprableMixin):
            _REPR_FIELDS = ('v',)

            def __init__(self, v):
                self.v = v

        self.assertEqual(repr(Test(2)), 'Test(2)')

        class Test2(ReprableMixin):
            _REPR_FIELDS = ('v',)
            _REPR_FULL_CLASSNAME = True

            def __init__(self, v):
                self.v = v

        self.assertTrue(repr(Test2(2)).endswith(
            'test_coding.test_structures.TestMisc.test_reprable_mixin.<locals>.Test2(2)'))

    def test_proxy(self):
        a = Proxy(5, wrap_operations=True)
        self.assertIsInstance(a + 5, Proxy)

        a = Proxy(5)
        self.assertNotIsInstance(a + 5, Proxy)
        self.assertIsInstance(a + 5, int)

        a = Proxy(5.5)
        self.assertEqual(math.floor(a), 5)

    def test_key_aware_defaultdict(self):
        a = KeyAwareDefaultDict(int)
        self.assertEqual(a['1'], 1)

    def test_dirty_dict(self):
        a = DirtyDict({1: 2, 3: 4})
        self.assertFalse(a.dirty)
        a[1] = 3
        self.assertTrue(a.dirty)
        a.clear_dirty()
        self.assertFalse(a.dirty)
        a[1] = 3
        self.assertFalse(a.dirty)
        del a[1]
        self.assertTrue(a.dirty)
        self.assertEqual(a.copy_and_clear_dirty(), {3: 4})
        a[3] = 5
        self.assertTrue(a.dirty)
        self.assertEqual(a.swap_and_clear_dirty(), {3: 5})
        self.assertFalse(a)
        self.assertFalse(a.dirty)

    def test_sliceable_deque(self):
        sd = SliceableDeque([1, 2, 3, 4, 5, 6, 7])
        self.assertEqual(list(sd[1:-1]), [2, 3, 4, 5, 6])
        self.assertEqual(sd.__getitem__(0), 1)

    def test_sorted_list(self):
        sl = SortedList([3, 2, 1], lambda a: a)
        self.assertEqual(sl[0], 1)
        self.assertEqual(sl[-1], 3)
        self.assertEqual(sl.index(2), 1)
        sl.add(5)
        self.assertEqual(sl.index(5), 3)
        self.assertRaises(ValueError, lambda: sl.index(6))
        sl.add(4)
        self.assertEqual(sl.index(5), 4)
        sl.remove(1)
        self.assertEqual(sl.index(5), 3)
        self.assertEqual(sl.pop(), 5)
        self.assertEqual(sl.popleft(), 2)

    def test_two_way_dict2(self):
        self.assertRaises(ValueError, lambda: TwoWayDictionary([(1, 2), (2, 2)]))

    def test_ranking(self):
        Entry = collections.namedtuple('Entry', ('a',))
        e1 = Entry(1)
        e2 = Entry(2)
        e3 = Entry(3)
        ranking = Ranking([e1, e2], lambda e: e.a)  # type: Ranking[Entry]

        self.assertEqual(ranking[0], e1)
        self.assertEqual(ranking[1], e2)
        self.assertEqual(ranking.get_position_of(e1), 0)
        self.assertEqual(list(ranking.get_sorted()), [e1, e2])

        ranking.add(e3)
        self.assertEqual(list(ranking.get_sorted()), [e1, e2, e3])
        ranking.remove(e1)
        self.assertEqual(list(ranking.get_sorted()), [e2, e3])
        self.assertEqual(ranking[-1], e3)
        e25 = Entry(2.5)
        ranking.add(e25)
        self.assertEqual(list(ranking.get_sorted()), [e2, e25, e3])

    def test_two_way_dict(self):
        twd = TwoWayDictionary({1: 2, 3: 4})
        self.assertEqual(twd.reverse[4], 3)
        del twd[1]
        self.assertRaises(KeyError, lambda: twd[1])
        self.assertRaises(KeyError, lambda: twd.reverse[2])
        twd[3] = 8
        self.assertEqual(twd.reverse[8], 3)
        self.assertRaises(KeyError, lambda: twd.reverse[4])

    def test_hashable_wrapper(self):
        class NotHashable:
            def __init__(self, a):
                self.a = a

            def __call__(self, *args, **kwargs):
                return 5

            def __hash__(self):
                raise TypeError()

        nh = NotHashable(5)
        nw = HashableWrapper(nh)
        self.assertEqual(nw.a, 5)
        nw.a = 4
        self.assertEqual(nw.a, 4)
        del nw.a
        self.assertRaises(AttributeError, lambda: getattr(nw, 'a'))
        self.assertEqual(nw, nh)
        self.assertEqual(nw, nw)
        hash(nw)
        self.assertEqual(nw(), 5)

    def test_dictionary_view(self):
        a = {1: 2, 3: 4}
        b = {4: 5, 6: 7}
        dva = DictionaryView(a, b)

        self.assertEqual(dva[1], 2)
        self.assertEqual(dva[4], 5)
        del dva[4]
        self.assertNotIn(4, b)
        self.assertNotIn(4, dva)
        dva[7] = 8
        self.assertIn(7, dva)
        self.assertIn(7, a)
        dva[6] = 10
        self.assertEqual(b[6], 10)

        dvb = DictionaryView(a, b, propagate_deletes=False, assign_to_same_dict=False)
        self.assertRaises(KeyError, lambda: dvb.__delitem__(6))
        dvb[6] = 11
        self.assertEqual(b[6], 10)
        self.assertEqual(a[6], 11)

        self.assertEqual(len(dvb), 4)

    def test_set_heap(self):
        a = SetHeap([1, 2, 3])
        self.assertIn(2, a)
        self.assertEqual(1, a.pop())
        self.assertNotIn(1, a)
        a.push_many([3, 4])
        self.assertEqual(len(a), 3)

    def test_frozendict(self):
        a = frozendict({1: 2, 3: 4})
        self.assertEqual(a[1], 2)
        self.assertRaises(TypeError, lambda: a.__setitem__(1, 3))
        b = frozendict({1: 2, 3: 4})
        self.assertEqual(a, b)
        self.assertEqual(hash(a), hash(b))
        self.assertRaises(TypeError, lambda: a.update({3: 5}))

    def continue_testing_omni(self, Omni):
        e1 = Omni(2)
        e2 = Omni(1)
        e3 = Omni(1)

        self.assertEqual(e2, e3)
        self.assertNotEqual(e1, e2)
        self.assertNotEqual(e1, e3)

        a = {
            e1: '1', e2: '2', e3: '3'
        }

        self.assertEqual(a[e1], '1')
        self.assertEqual(hash(e1), hash(2))

    def test_omni_not_filled_in_fields(self):
        class Omni1(OmniHashableMixin):
            _HASH_FIELDS_TO_USE = 'a'

            def __init__(self, a=None):
                if a is not None:
                    self.a = a

        self.assertNotEqual(Omni1(3), Omni1())
        self.assertFalse(Omni1() == Omni1(3))

    def test_omni_different_hierarchies(self):
        class Omni1(OmniHashableMixin):
            _HASH_FIELDS_TO_USE = 'a'

            def __init__(self, a):
                self.a = a

        class Omni2(OmniHashableMixin):
            _HASH_FIELDS_TO_USE = 'a'

            def __init__(self, a):
                self.a = a

        self.assertNotEqual(Omni1(1), Omni2(1))
        self.assertFalse(Omni1(1) == Omni2(1))

    def test_omni_single_field(self):
        class Omni(OmniHashableMixin):
            _HASH_FIELDS_TO_USE = 'a'

            def __init__(self, a):
                self.a = a

        self.continue_testing_omni(Omni)

    def test_omni(self):
        class Omni(OmniHashableMixin):
            _HASH_FIELDS_TO_USE = ('a', )

            def __init__(self, a):
                self.a = a
        self.continue_testing_omni(Omni)

    def test_tbsh(self):
        tbh = TimeBasedSetHeap()

        tbh.put(10, 'ala')
        tbh.put(20, 'ma')
        tbh.put(30, 'kota')
        tbh.put(30, 'ala')
        self.assertIn((30, 'ala'), tbh)
        self.assertNotIn((10, 'ala'), tbh)

        q = set(tbh.pop_less_than(25))

        self.assertIn((20, 'ma'), q)
        self.assertNotIn((30, 'ala'), q)
        self.assertNotIn((30, 'kota'), q)

        self.assertIsInstance(copy.copy(tbh), TimeBasedSetHeap)
        self.assertIsInstance(copy.deepcopy(tbh), TimeBasedSetHeap)

        item = tbh.pop_timestamp(30)
        self.assertTrue(item == 'kota' or item == 'ala')

    def test_tbh(self):
        tbh = TimeBasedHeap()

        tbh.put(10, 'ala')
        tbh.put(20, 'ma')
        tbh.put(30, 'kota')

        self.assertEqual(tbh.peek_closest(), (10, 'ala'))

        q = set(tbh.pop_less_than(25))

        self.assertIn((10, 'ala'), q)
        self.assertIn((20, 'ma'), q)
        self.assertNotIn((30, 'kota'), q)

        self.assertIsInstance(copy.copy(tbh), TimeBasedHeap)
        self.assertIsInstance(copy.deepcopy(tbh), TimeBasedHeap)

        item = tbh.pop_timestamp(30)
        self.assertEqual(item, 'kota')
        self.assertNotIn((30, 'kota'), tbh)

    def test_improve(self):
        tbh = TimeBasedHeap()
        tbh.put(10, 'ala')

        self.assertIn('ala', list(tbh.items()))

    def test_typednamedtuple(self):
        tnt = typednamedtuple('tnt',
                              ('x', float),
                              ('y', float))

        a = tnt('2.5', y=3)

        self.assertEqual(a.x, 2.5)
        self.assertEqual(a.y, 3.0)
        self.assertIsInstance(a.y, float)
        self.assertIsInstance(a, tuple)

    def test_def(self):
        clk = mock.Mock(return_value=0)

        tbh = TimeBasedHeap(default_clock_source=clk)
        tbh.put('ala')
        tbh.put(4, 'alla')

        clk.return_value = 3

        q = set(tbh.pop_less_than())

        self.assertIn((0, 'ala'), q)
        self.assertNotIn((4, 'alla'), q)

    def test_foobar(self):
        tbh = TimeBasedHeap()
        bytes(tbh)
        str(tbh)
        repr(tbh)
        copy.copy(tbh)
        copy.deepcopy(tbh)


class TestDictObject(unittest.TestCase):

    def test_constructor(self):
        a = DictObject({'a': 5})
        b = apply_dict_object(a)
        self.assertIs(a, b)

    def test_valid_schema(self):
        self.assertTrue(DictObject({'a': 5, 'b': 'test'}).is_valid_schema({'a': int, 'b': str}))
        self.assertFalse(DictObject({'a': 5, 'b': 'test'}).is_valid_schema({'a': int, 'b': int}))
        self.assertTrue(DictObject({'a': 5, 'b': 'test'}).is_valid_schema(a=int, b=str))
        self.assertFalse(DictObject({'a': 5, 'b': 'test'}).is_valid_schema(a=int, b=int))

    def test_apply_dict_object(self):
        a = {'a': {'b': 5}, 'c': [{'a': 5}]}
        b = DictObject({'a': DictObject({'b': 5}), 'c': [DictObject({'a': 5})]})
        a = apply_dict_object(a)
        self.assertEqual(apply_dict_object(a), b)
        self.assertIsInstance(a, DictObject)
        self.assertIsInstance(a.a, DictObject)
        self.assertIsInstance(a.c[0], DictObject)

    def test_dict_object(self):
        a = {'a': 1, 'b': 2}

        a = DictObject(a)

        self.assertEqual(a.a, 1)
        self.assertEqual(a.b, 2)
        self.assertRaises(AttributeError, lambda: a.c)
        del a.a
        self.assertRaises(AttributeError, lambda: a.a)
        a.a = 5
        self.assertEqual(a.a, 5)

        def delete():
            del a.c

        self.assertRaises(AttributeError, delete)

    def test_copying(self):
        a = DictObject({1: 2})
        b = copy.copy(a)
        self.assertEqual(a, b)
        self.assertFalse(id(a) == id(b))
        c = copy.deepcopy(a)
        c[1] = 3
        self.assertEqual(a[1], 2)
        self.assertEqual(c[1], 3)


class TestHeap(unittest.TestCase):
    def test_push(self):
        tbh = Heap()
        tbh.push(10, 'A')
        self.assertEqual((10, 'A'), tbh.pop())
        tbh.push((10, 'A'))
        self.assertEqual(tbh[0], (10, 'A'))
        self.assertEqual((10, 'A'), tbh.pop())
        self.assertEqual(tbh.data, [])

    def test_tbh_iter(self):
        tbh = Heap()

        tb = [(10, 'ala'), (20, 'ma'), (30, 'kota'), (5, 'yo')]

        tbh.push_many(tb)
        self.assertEqual(sorted(tb), list(tbh.iter_ascending()))
        self.assertEqual(sorted(tb, reverse=True), list(tbh.iter_descending()))

    def test_tbh(self):
        tbh = Heap()

        tbh.push_many([
            (10, 'ala'),
            (20, 'ma')
        ])

        self.assertIn((10, 'ala'), tbh)
        self.assertIn((20, 'ma'), tbh)

        tbh.filter_map(filter_fun=lambda x: x[0] != 20,
                       map_fun=lambda x: (x[0] + 10, 'test3'))

        self.assertIn((20, 'test3'), tbh)
        self.assertNotIn((10, 'ala'), tbh)
        self.assertNotIn((20, 'ma'), tbh)

        item = tbh.pop_item((20, 'test3'))
        self.assertNotIn((20, 'test3'), tbh)
        self.assertEqual(item, (20, 'test3'))


class TestImmutable(unittest.TestCase):
    def _test_an_instance(self, a):
        self.assertEqual(a.x, 2.5)
        self.assertEqual(a.y, 2)

        def change_x():
            a.x = 2

        self.assertRaises(TypeError, change_x)

        def delete_x():
            del a.x

        self.assertRaises(TypeError, delete_x)

    def test_immutable(self):
        class Point2D(Immutable):
            def __init__(self, x: float, y: float):
                self.x = x
                self.y = y

        self._test_an_instance(Point2D(2.5, 2))

    def test_immutable_abc(self):
        class Point2D(abc.ABC, Immutable):
            def __init__(self, x: float, y: float):
                self.x = x
                self.y = y

        self._test_an_instance(Point2D(2.5, 2))

    def test_advanced_inheritance_hierarchy(self):
        class AnObject:
            pass

        class BaseClass(metaclass=abc.ABCMeta):
            pass

        class AnotherBaseClass(metaclass=abc.ABCMeta):
            pass

        class TwinnedBaseClass(AnotherBaseClass, metaclass=abc.ABCMeta):
            pass

        class Point2D:
            def __init__(self, x: float, y: float):
                self.x = x
                self.y = y

        class Point3D(Point2D, BaseClass, TwinnedBaseClass, AnObject, Immutable):
            pass

        self._test_an_instance(Point3D(2.5, 2))



import os
import threading
import unittest

from satella.coding import silence_excs


class ExceptionHandlingTestCase(unittest.TestCase):
    def setUp(self):
        try:
            os.environ
        except AttributeError:
            os.environ = {}

    def tearDown(self):
        with silence_excs(ValueError):
            self.exception_handler.uninstall()

    def make_exception(self, fun=lambda: {}['wtf']):

        class Hello(threading.Thread):
            def __init__(self):
                super(Hello, self).__init__()

            def run(self):
                return fun()

        p = Hello()
        p.start()
        p.join()



import itertools
import unittest

from satella.coding.sequences import choose, choose_one, infinite_counter, take_n, is_instance, \
    is_last, add_next, half_cartesian, skip_first, zip_shifted, stop_after, group_quantity, \
    iter_dict_of_list, shift, other_sequence_no_longer_than, count, even, odd, Multirun, n_th, \
    unique, length, map_list, smart_zip, is_empty, make_list, RollingArithmeticAverage, \
    enumerate2, infinite_iterator, to_iterator, filter_out_false, filter_out_nones, \
    index_of, index_of_max, try_close, f_range, iterate_callable, AlreadySeen, \
    append_sequence
from satella.coding.predicates import x


class TestSequences(unittest.TestCase):

    def test_append_sequence(self):
        a = [(1,), (2,), (3,)]
        self.assertEqual(list(append_sequence(a, 1, 2)), [(1, 1, 2), (2, 1, 2), (3, 1, 2)])

    def test_already_seen(self):
        as_ = AlreadySeen()
        self.assertTrue(as_.is_unique(1))
        self.assertFalse(as_.is_unique(1))

    def test_iterate_callable(self):
        a = [1, 2, 3]
        b = lambda i: a[i]
        self.assertEqual(list(iterate_callable(b)), [1, 2, 3])

    def test_f_range(self):
        self.assertRaises(TypeError, f_range)
        self.assertRaises(TypeError, lambda: f_range(1, 2, 3, 4))
        self.assertEqual(list(f_range(0.5, 1.5, 0.5)), [0.5, 1.0])
        self.assertEqual(list(f_range(0.5, -1.0, -0.5)), [0.5, 0.0, -0.5])

    def test_try_close(self):
        def iterator():
            yield 2

        a = iterator()
        try_close(a)
        try_close(iter([1, 2, 3]))

    def test_index_of_max(self):
        self.assertRaises(ValueError, lambda: index_of_max([]))
        self.assertEqual(index_of_max([1, 3, 2]), 1)

    def test_index_of(self):
        self.assertEqual(index_of(x == 2, [1, 2, 3]), 1)
        self.assertRaises(ValueError, lambda: index_of(x == 3, [1, 2, 4]))

    def test_filter_out_false(self):
        a = [0, 1, 2, 3, 4, False, None]
        self.assertEqual(filter_out_false(a), [1, 2, 3, 4])

    def test_filter_out_nones(self):
        a = [0, 1, 2, 3, 4, False, None]
        self.assertEqual(filter_out_nones(a), [0, 1, 2, 3, 4, False])

    def test_to_iterator(self):
        @to_iterator
        def is_even(x):
            return not (x % 2)

        self.assertEqual(list(is_even([1, 2, 3, 4])), [False, True, False, True])

    def test_infinite_iterator(self):
        if1 = infinite_iterator(1)
        for _ in range(100):
            self.assertEqual(next(if1), 1)
        if2 = infinite_iterator(return_factory=lambda: 1)
        for _ in range(100):
            self.assertEqual(next(if1), 1)

    def test_enumerate2(self):
        a = [1, 2, 3, 4]
        self.assertEqual(list(enumerate2(a, -1, -1)), [(-1, 1), (-2, 2), (-3, 3), (-4, 4)])

    def test_rolling_arith_avg(self):
        a = RollingArithmeticAverage(10)
        for i in range(10):
            a.insert(1)
        self.assertEqual(a.avg(), 1)
        a.insert(5)
        self.assertGreaterEqual(a.avg(), 1)
        self.assertLessEqual(len(a.queue), 10)
        self.assertEqual(a.tot_sum, 9+5)

    def test_make_list(self):
        a = [1, 2, 3]
        b = make_list(a, 3)
        self.assertEqual(len(b), 3)
        b[0][2] = 4
        self.assertNotEqual(b[1][2], 4)

    def test_make_list_deep_copy(self):
        class A:
            def __init__(self, c):
                self.a = c
        a = A(3)
        b = make_list(a, 3, deep_copy=True)
        self.assertEqual(len(b), 3)
        b[0].a = 4
        self.assertNotEqual(a.a, 4)

    def test_empty(self):
        def empty_iterator():
            if False:
                yield 2

        e = empty_iterator()
        self.assertTrue(is_empty(e))
        self.assertFalse(is_empty([1]))

    def test_smart_zip(self):
        a1 = [(1, 1), (1, 2), (1, 3)]
        a2 = [1, 2, 3]
        a = list(smart_zip(a1, a2))
        self.assertEqual(a, [(1, 1, 1), (1, 2, 2), (1, 3, 3)])

    def test_map_list(self):
        _list = map_list(x*2, range(5))
        self.assertEqual(_list, [0, 2, 4, 6, 8])

    def test_length(self):
        def iterator():
            i = 0
            for i in range(10):
                if i == 3:
                    return
                yield i
        self.assertEqual(length(iterator()), 3)

    def test_unique(self):
        a = [1, 2, 3, 1, 2, 3]
        self.assertEqual(list(unique(a)), [1, 2, 3])

    def test_n_th(self):
        self.assertEqual(n_th(itertools.count()), 0)
        self.assertEqual(n_th(itertools.count(), 3), 3)

    def test_multirun_empty(self):
        a = Multirun([])
        a += 3
        self.assertFalse(a.sequence)

    def test_multirun(self):
        class Counter:
            def __init__(self, value=0):
                self.value = value

            def add(self, value):
                self.value += value

            def __iadd__(self, other):
                self.add(other)

            def __eq__(self, other):
                return self.value == other.value

        a = [Counter(2), Counter(3)]

        b = Multirun(a)
        self.assertTrue(b)
        b.add(3)
        self.assertEqual(a, [Counter(5), Counter(6)])

        b += 3
        self.assertEqual(a, [Counter(8), Counter(9)])

    def test_even_and_odd(self):
        a = [0, 1, 2, 3, 4, 5, 6]
        self.assertEqual(list(even(a)), [0, 2, 4, 6])
        self.assertEqual(list(odd(a)), [1, 3, 5])

    def test_count(self):
        self.assertEqual(list(count([None, None, None], 5, -2)), [5, 3, 1])

    def test_count_without_start(self):
        self.assertEqual(list(count([None, None, None])), [0, 1, 2])

    def test_other_sequence_no_longer_than(self):
        s1 = [1, 2, 3]
        s2 = [3, 4, 5, 6]
        self.assertEqual(list(other_sequence_no_longer_than(s1, s2)), [3, 4, 5])

    def test_shift(self):
        self.assertEqual(list(shift([1, 2, 3], 1)), [2, 3, 1])
        self.assertEqual(list(shift([1, 2, 3], -1)), [3, 1, 2])

    def test_iter_dict_of_list(self):
        a = {1: [1, 2, 3]}
        self.assertEqual(list(iter_dict_of_list(a)), [(1, 1), (1, 2), (1, 3)])

    def test_group_quantity(self):
        self.assertEqual(list(group_quantity(3, [1, 2, 3, 4, 5, 6, 7, 8, 9, 10])),
                         [[1, 2, 3], [4, 5, 6], [7, 8, 9], [10]])
        self.assertEqual(list(group_quantity(3, 'ala ma kota')), ['ala', ' ma', ' ko', 'ta'])

    def test_group_quantity_range(self):
        def test_iter():
            yield from range(10)

        self.assertEqual(list(group_quantity(3, test_iter())),
                         [[0, 1, 2], [3, 4, 5], [6, 7, 8], [9]])
        self.assertEqual(list(group_quantity(3, range(10))),
                         [[0, 1, 2], [3, 4, 5], [6, 7, 8], [9]])

    def test_stop_after(self):
        a = [1, 2, 3, 4, 5, 6, 7]
        a = list(stop_after(a, 2))
        self.assertEqual([1, 2], a)

    def test_zip_shifted(self):
        a = list(zip_shifted(([1, 2, 3, 4], 1), [1, 2, 3, 4]))
        self.assertEqual(a, [(2, 1), (3, 2), (4, 3), (1, 4)])

    def test_zip_shifted_negative(self):
        a = list(zip_shifted(([1, 2, 3, 4], -1), [1, 2, 3, 4]))
        self.assertEqual(a, [(4, 1), (1, 2), (2, 3), (3, 4)])

    def test_skip_first(self):
        a = [1, 2, 3, 4, 5]
        b = list(skip_first(a, 1))
        self.assertEqual(b, [2, 3, 4, 5])

    def test_half_cartesian(self):
        a = set(half_cartesian([1, 2, 3]))
        b = {(1, 1), (1, 2), (1, 3), (2, 2), (2, 3), (3, 3)}
        self.assertEqual(a, b)

    def test_half_cartesian_include_same_pairs_false(self):
        a = set(half_cartesian([1, 2, 3], include_same_pairs=False))
        b = {(1, 2), (1, 3), (2, 3)}
        self.assertEqual(a, b)

    def test_add_next(self):
        self.assertEqual(list(add_next([1, 2, 3, 4, 5])),
                         [(1, 2), (2, 3), (3, 4), (4, 5), (5, None)])
        self.assertEqual(list(add_next([1])), [(1, None)])
        self.assertEqual(list(add_next([])), [])
        self.assertEqual(list(add_next([1, 2, 3, 4, 5], skip_last=True)),
                         [(1, 2), (2, 3), (3, 4), (4, 5)])

    def test_add_next_wrap_over(self):
        self.assertEqual(list(add_next([1, 2, 3, 4, 5], True)),
                         [(1, 2), (2, 3), (3, 4), (4, 5), (5, 1)])

    def test_is_last(self):
        for is_last_flag, elem in is_last([1, 2, 3, 4, 5]):
            self.assertTrue(not is_last_flag ^ (elem == 5))

    def test_take_n(self):
        subset = take_n(infinite_counter(), 10)
        for a, b in zip(range(10), subset):
            self.assertEqual(a, b)

    def test_infinite_counter(self):
        p = infinite_counter(1)
        for i in range(1, 10):
            a = next(p)
            self.assertEqual(a, i)

    def test_choose_one(self):
        self.assertEqual(choose_one(x == 2, [1, 2, 3, 4, 5]), 2)
        self.assertEqual(choose(x % 2 == 0, [1, 2, 3, 4, 5]), 2)
        self.assertRaises(ValueError, lambda: choose_one(lambda x: x % 2 == 0, [1, 2, 3, 4, 5]))
        self.assertRaises(ValueError, lambda: choose_one(lambda x: x == 0, [1, 2, 3, 4, 5]))

    def test_is_instance(self):
        objects = [object(), object(), [], [], object()]
        self.assertEqual(len(list(filter(is_instance(list), objects))), 2)



import bisect
import typing as tp
import unittest

from satella.coding.predicates import x
from satella.coding.sequences import index_of
from satella.coding.structures import DBStorage, SyncableDroppable
from satella.coding.typing import K, KVTuple, V


class Iterator:
    def __init__(self, db_storage, it: tp.Iterator):
        self.it = it
        self.db_storage = db_storage

    def __iter__(self):
        return self.it

    def __next__(self):
        return next(self.it)

    def close(self):
        self.db_storage.iterators -= 1


class MyDBStorage(DBStorage):
    def delete(self, key: K) -> None:
        index = index_of(x[0] == key, self.data)
        del self.data[index]

    def on_change_synced_up_to(self, synced_up_to: tp.Optional[K]) -> None:
        self.synced = synced_up_to

    def on_change_stop_entry(self, stop_entry: tp.Optional[K]) -> None:
        self.stopped = stop_entry

    def iterate(self, starting_key: tp.Optional[K]) -> tp.Iterator[KVTuple]:
        if starting_key is None:
            it = iter(self.data)
        else:
            index = bisect.bisect_left([y[0] for y in self.data], starting_key)
            it = iter(self.data[index:])
        self.iterators += 1
        return Iterator(self, it)

    def __init__(self):
        self.data = []          # type: tp.List[KVtuple]
        self.started = None
        self.stopped = None
        self.synced = None
        self.iterators = 0

    def on_change_start_entry(self, start_entry: tp.Optional[K]) -> None:
        self.started = start_entry

    def put(self, key: K, value: V) -> None:
        self.data.append((key, value))


class TestSyncableDroppable(unittest.TestCase):
    def test_sync_drop(self):
        db = MyDBStorage()
        sd = SyncableDroppable(db, None, None, None, 100, 200)
        sd.on_new_data(100, 5)
        sd.on_new_data(150, 5)
        sd.on_new_data(180, 5)
        sd.on_new_data(200, 5)
        sd.on_new_data(220, 5)
        self.assertEqual(sd.first_key_in_memory, 100)
        sd.cleanup_keep_in_memory()
        self.assertEqual(sd.stop_entry, 220)
        self.assertEqual(sd.start_entry, 100)
        self.assertEqual(sd.first_key_in_memory, 150)
        self.assertEqual(db.data, [(100, 5)])
        sd.on_new_data(305, 5)
        sd.on_new_data(400, 5)
        sd.cleanup_keep_in_db()
        self.assertEqual(sd.start_entry, 200)
        self.assertEqual(db.data, [])

        sd.cleanup_keep_in_memory()
        self.assertEqual(db.data, [(200, 5), (220, 5)])
        synces = list(sd.on_sync_request())
        self.assertEqual(synces, [(200, 5), (220, 5), (305, 5), (400, 5)])
        synced_up_to = synces[-1][0]
        sd.on_synced_up_to(synced_up_to)
        self.assertRaises(ValueError, sd.on_sync_request)
        sd.on_new_data(405, 5)
        sd.on_new_data(409, 5)
        synces = list(sd.on_sync_request())
        self.assertEqual(synces, [(405, 5), (409, 5)])
        sd.on_new_data(510, 5)
        sd.cleanup_keep_in_memory()
        self.assertEqual(db.data, [(200, 5), (220, 5), (305, 5), (400, 5), (405, 5), (409, 5)])
        self.assertEqual(list(sd.get_archive(220, 420)), [(220, 5), (305, 5), (400, 5), (405, 5), (409, 5)])
        sd.cleanup_keep_in_db()
        self.assertEqual(db.data, [(400, 5), (405, 5), (409, 5)])
        self.assertFalse(db.iterators)
        self.assertEqual(sd.get_latest_value(), (510, 5))



import unittest

from satella.coding.typing import ExceptionClassType


class TestTyping(unittest.TestCase):
    def test_exception_type(self):
        def catch_exception(e) -> ExceptionClassType:
            try:
                e()
            except Exception as e:
                return e.__class__

        a = catch_exception(lambda: hello)



import typing as tp
from abc import abstractmethod, ABCMeta

from satella.coding.concurrent import IntervalTerminableThread
from .collector import sleep_cpu_aware, CPUProfileBuilderThread
from satella.time import measure, parse_time_string


class CPUTimeAwareIntervalTerminableThread(IntervalTerminableThread, metaclass=ABCMeta):
    """
    An IntervalTerminableThread that can call the loop a bit faster than usual,
    based of current CPU time metrics.

    :param seconds: time that a single looping through should take. This will
        include the time spent on calling .loop(), the rest of this time will
        be spent safe_sleep()ing.
        Can be alternatively a time string
    :param max_sooner: amount of seconds that is ok to call this earlier.
        Default is 10% seconds.
    :param percentile: percentile that CPU usage has to fall below to call it earlier.
    :param wakeup_interval: amount of seconds to wake up between to check for _terminating status.
        Can be also a time string

    Same concerns for :code:`terminate_on` as in
    :class:`~satella.coding.concurrent.TerminableThread` apply.
    """

    def __init__(self, seconds: tp.Union[str, float], max_sooner: tp.Optional[float] = None,
                 percentile: float = 0.3,
                 wakeup_interval: tp.Union[str, float] = '3s', *args, **kwargs):
        self.seconds = parse_time_string(seconds)
        self.wakeup_interval = parse_time_string(wakeup_interval)
        self.max_sooner = max_sooner or seconds * 0.1
        cp_bt = CPUProfileBuilderThread()
        cp_bt.request_percentile(percentile)
        self.percentile = percentile
        super().__init__(seconds, *args, **kwargs)

    @abstractmethod
    def loop(self) -> None:
        """
        Override me!
        """

    def _execute_measured(self) -> float:
        with measure() as measurement:
            self.loop()
        return measurement()

    def __sleep_waiting_for_cpu(self, how_long: float) -> None:
        cp_bt = CPUProfileBuilderThread()
        per_val = cp_bt.percentile(self.percentile)

        while how_long > 0 and not self._terminating:
            time_to_sleep = min(self.wakeup_interval, how_long)

            if sleep_cpu_aware(time_to_sleep, per_val):
                break
            how_long -= time_to_sleep

    def __sleep(self, how_long: float) -> None:
        if how_long > self.max_sooner:
            if self.safe_sleep(how_long - self.max_sooner):
                return
            how_long = self.max_sooner
        self.__sleep_waiting_for_cpu(how_long)

    def run(self):
        try:
            try:
                self.prepare()
            except Exception as e:
                if self._terminate_on is not None:
                    if isinstance(e, self._terminate_on):
                        self.terminate()
                        return
                raise
            while not self.terminating:
                try:
                    measured = self._execute_measured()
                except Exception as e:
                    if self._terminate_on is not None:
                        if isinstance(e, self._terminate_on):
                            self.terminate()
                            return
                    raise
                seconds_to_wait = self.seconds - measured
                if seconds_to_wait > 0:
                    self.__sleep(seconds_to_wait)
                elif seconds_to_wait < 0:
                    self.on_overrun(measured)
        finally:
            self.cleanup()



import typing as tp
import logging
import threading
import time

from satella.coding.recast_exceptions import log_exceptions
from .monitor import Monitor
from ..structures.heaps.time import TimeBasedHeap
from ..structures.singleton import Singleton
from ...time.parse import parse_time_string

logger = logging.getLogger(__name__)


class Timer:
    """
    A copy of threading.Timer but all objects are backed and waited upon in a single thread.
    They can be executed either in background monitor's thread or a separate thread can be
    spawned for them.

    There might be up to a second of delay before the timer is picked up.

    If spawn_separate is False, exceptions will be logged

    :param interval: amount of seconds that should elapsed between calling start() and function
        executing. Can be also a time string.
    :param function: function to execute
    :param args: argument for function
    :param kwargs: kwargs for function
    :param spawn_separate: whether to call the function in a separate thread
    """
    __slots__ = ('args', 'kwargs', 'spawn_separate', 'interval',
                 'function', 'execute_at', 'cancelled')

    def __init__(self, interval: tp.Union[str, float], function, args=None, kwargs=None,
                 spawn_separate=False):
        self.args = args or []
        self.kwargs = kwargs or {}
        self.spawn_separate = spawn_separate
        self.interval = parse_time_string(interval)
        self.function = function
        self.execute_at = None
        self.cancelled = False

    def start(self) -> None:
        """
        Order this timer task to be executed in interval seconds
        """
        execute_at = time.monotonic() + self.interval
        tbt = TimerBackgroundThread()
        with Monitor.acquire(tbt):
            tbt.timer_objects.put(execute_at, self)

    def cancel(self) -> None:
        """Do not execute this timer"""
        self.cancelled = True

    def _try_execute(self):
        if self.cancelled:
            return
        if self.spawn_separate:
            threading.Thread(target=self.function, args=self.args, kwargs=self.kwargs,
                             daemon=True).start()
        else:
            self.function(*self.args, **self.kwargs)


@Singleton
class TimerBackgroundThread(threading.Thread, Monitor):
    def __init__(self):
        super().__init__(name='timer background thread', daemon=True)
        Monitor.__init__(self)
        self.timer_objects = TimeBasedHeap()  # type: TimeBasedHeap[Timer]
        self.start()

    def run(self):
        while True:
            try:
                ts, _ = self.timer_objects.peek_closest()
                delay = min(ts - time.monotonic(), 1)
                if delay < 0:
                    delay = 0
            except IndexError:
                delay = 1
            time.sleep(delay)
            with Monitor.acquire(self):
                items_to_exec = self.timer_objects.pop_less_than(time.monotonic())

            for item in items_to_exec:
                with log_exceptions(logger, swallow_exception=True):
                    item[1]._try_execute()



import time
import typing as tp
from concurrent.futures import wait, ThreadPoolExecutor

from .atomic import AtomicNumber
from .futures import ExecutorWrapper
from .thread import Condition
from ...exceptions import WouldWaitMore
from ...time.measure import measure


def sync_threadpool(tpe: tp.Union[ExecutorWrapper, ThreadPoolExecutor],
                    max_wait: tp.Optional[float] = None) -> None:
    """
    Make sure that every thread of given thread pool executor is done processing
    jobs scheduled until this moment.

    Make sure that other tasks do not submit anything to this thread pool executor.

    :param tpe: thread pool executor to sync. Can be also a ExecutorWrapper.
    :param max_wait: maximum time to wait. Default, None, means wait forever
    :raises WouldWaitMore: timeout exceeded. Raised only when max_wait is not None.
    """
    if isinstance(tpe, ExecutorWrapper):
        return sync_threadpool(tpe.executor, max_wait=max_wait)

    assert isinstance(tpe, ThreadPoolExecutor), 'Must be a ThreadPoolExecutor!'

    with measure(timeout=max_wait) as measurement:
        # noinspection PyProtectedMember
        workers = tpe._max_workers
        atm_n = AtomicNumber(workers)
        cond = Condition()

        def decrease_atm():
            nonlocal atm_n
            atm_n -= 1
            cond.wait()

        futures = [tpe.submit(decrease_atm) for _ in range(workers)]

        # wait for all currently scheduled jobs to be picked up
        # noinspection PyProtectedMember
        while tpe._work_queue.qsize() > 0:
            if max_wait is not None:
                if measurement() > max_wait:
                    for future in futures:
                        future.cancel()
                    raise WouldWaitMore('timeout exceeded')
            time.sleep(0.5)

        if max_wait is None:
            atm_n.wait_until_equal(0)
        else:
            while measurement() < max_wait:
                try:
                    atm_n.wait_until_equal(0, 1)
                    break
                except WouldWaitMore:
                    continue
            else:
                raise WouldWaitMore('timeout exceeded')
        cond.notify_all()
        wait(futures)



import inspect
import logging
import time
import unittest

from satella.coding.sequences import n_th

from satella.exceptions import MetricAlreadyExists
from satella.instrumentation.metrics import getMetric, MetricLevel, MetricData, \
    MetricDataCollection, AggregateMetric, LabeledMetric

logger = logging.getLogger(__name__)


def choose(postfix: str, mdc: MetricDataCollection, labels=None) -> MetricData:
    for child in mdc.values:
        if child.name.endswith(postfix) and child.labels == (labels or {}):
            return child


class TestMetric(unittest.TestCase):

    def test_child_metrics(self):
        """A RUNTIME parent does not display DEBUG child's metric"""
        parent = getMetric('test.child.metric', metric_level=MetricLevel.RUNTIME)
        child = getMetric('test.child.metric.kid', 'int', metric_level=MetricLevel.DEBUG)
        child.runtime(2)
        data = parent.to_metric_data()
        kid = choose('.kid', data)
        self.assertIsNone(kid)

    def test_default_float(self):
        met = getMetric('test.float', 'float')
        data = met.to_metric_data()
        v = choose('', data)
        self.assertEqual(v.value, 0.0)

    def test_uptime_metric(self):
        up_metric = getMetric('uptime.metric', 'uptime')
        time.sleep(1)
        self.assertGreaterEqual(n_th(up_metric.to_metric_data().values).value, 1)

    def test_forbidden_characters(self):
        self.assertRaises(ValueError, lambda: getMetric('forbidden-characters', 'int'))

    def test_labeled_metric_with_callable(self):
        m1 = getMetric('lmc1', 'callable')
        lm = LabeledMetric(m1, key='value')
        lm.callable = lambda: 0

        self.assertEqual(choose('lmc1', m1.to_metric_data(), {'key': 'value'}).value, 0)

    def test_aggregate(self):
        m1 = getMetric('am1', 'summary')
        m2 = getMetric('am2', 'histogram')
        am = AggregateMetric(m1, m2)
        am.runtime(2)
        data = choose('.count', m1.to_metric_data())
        self.assertEqual(data.value, 1)
        data = choose('.count', m2.to_metric_data())
        self.assertEqual(data.value, 1)

    def test_callable(self):
        callable_ = getMetric('callable', 'callable', value_getter=lambda: 5.0)
        self.assertEqual(list(callable_.to_metric_data().values)[0].value, 5.0)

    def test_linkfail(self):
        d = {'online': False, 'offline': False}

        def on_online(address, labels):
            d['online'] = True

        def on_offline(address, labels):
            d['offline'] = True

        linkfail = getMetric('linkfail', 'linkfail', consecutive_failures_to_offline=10,
                             consecutive_successes_to_online=10, callback_on_online=on_online,
                             callback_on_offline=on_offline)

        for i in range(10):
            linkfail.runtime(False)

        self.assertEqual(choose('.status', linkfail.to_metric_data()).value, 0)

        for i in range(10):
            linkfail.runtime(True)

        self.assertTrue(d['online'])
        self.assertTrue(d['offline'])

        self.assertEqual(choose('.status', linkfail.to_metric_data()).value, 1)

    def setUp(self) -> None:
        getMetric('').reset()

    def test_embedded_submetrics_labels_getter(self):
        metric = getMetric('embedded_submetrics_test', 'int')
        metric.runtime(1, label='value')
        self.assertTrue(metric.get_specific_metric_data({'label': 'value'}).values)

    def test_histogram(self):
        metric = getMetric('test_histogram', 'histogram')
        metric.runtime(1)
        metric.runtime(2.6)
        metric_data = metric.to_metric_data()
        self.assertEqual(choose('', metric_data, {'le': 2.5, 'ge': 1.0}).value, 1)
        self.assertEqual(choose('sum', metric_data).value, 3.6)
        self.assertEqual(choose('count', metric_data).value, 2)

    def test_histogram_children(self):
        metric = getMetric('test_histogram', 'histogram')
        metric.runtime(1, label='value')
        metric.runtime(2.6, label='value')
        metric_data = metric.to_metric_data()
        self.assertEqual(choose('', metric_data, {'le': 2.5, 'ge': 1.0, 'label': 'value'}).value, 1)
        self.assertEqual(choose('sum', metric_data, {'label': 'value'}).value, 3.6)
        self.assertEqual(choose('count', metric_data, {'label': 'value'}).value, 2)
        self.assertEqual(choose('total', metric_data, {'le': 2.5, 'ge': 1.0}).value, 1)
        self.assertEqual(choose('total.sum', metric_data).value, 3.6)
        self.assertEqual(choose('total.count', metric_data).value, 2)

    def test_empty(self):
        metric = getMetric('empty', 'empty')
        self.assertEqual(len(metric.to_metric_data().values), 0)

    def test_metric_already_exists(self):
        getMetric('testmetric2', 'cps')
        try:
            getMetric('testmetric2', 'summary')
        except MetricAlreadyExists as e:
            self.assertEqual(e.name, 'testmetric2')
            self.assertEqual(e.requested_type, 'summary')
            self.assertEqual(e.existing_type, 'cps')
        else:
            self.fail('Exception not raised')

    def test_int_children(self):
        metric = getMetric('test_children_int', 'int')
        metric.runtime(1, label='value')
        should_be_equal_to = MetricDataCollection(
            MetricData('test_children_int', 1, {'label': 'value'}))
        self.assertTrue(metric.to_metric_data().strict_eq(should_be_equal_to))

    def test_quantile_measure_generator(self):
        metric = getMetric('my_metric', 'summary', quantiles=[0.5])

        @metric.measure()
        def generator():
            yield 2
            time.sleep(1)
            yield 3

        for v in generator():
            pass
        self.assertTrue(inspect.isgeneratorfunction(generator))
        self.assertGreaterEqual(next(iter(metric.to_metric_data().values)).value, 1)

    def test_aggregate_metric_measure_generator(self):
        my_metric = getMetric('my_metric', 'summary', quantiles=[0.5])
        metric = AggregateMetric(my_metric)

        @metric.measure()
        def generator():
            yield 2
            time.sleep(1)
            yield 3

        for _ in generator():
            pass
        self.assertTrue(inspect.isgeneratorfunction(generator))
        self.assertGreaterEqual(next(iter(my_metric.to_metric_data().values)).value, 1)

    def test_quantile_context_manager(self):
        metric = getMetric('test_metric', 'summary', quantiles=[0.5])
        with metric.measure():
            time.sleep(1)
        self.assertGreaterEqual(next(iter(metric.to_metric_data().values)).value, 1)

    def test_counter_measure(self):
        metric = getMetric('test2', 'counter')
        with metric.measure():
            time.sleep(2)
        self.assertGreaterEqual(next(iter(metric.to_metric_data().values)).value, 1)

    def test_quantile_children(self):
        metric = getMetric('my_metric', 'summary', quantiles=[0.5], enable_timestamp=True)
        metric.runtime(10.0, label='value')
        metric.runtime(20.0, label='wtf')
        metr = metric.to_metric_data()
        self.assertEqual(choose('sum', metr).value, 30.0)
        self.assertEqual(choose('count', metr).value, 2)
        self.assertEqual(choose('total', metr, {'quantile': 0.5}).value, 15.0)
        self.assertTrue(all(x.timestamp is not None for x in metr.values))

    def test_quantile(self):
        metric = getMetric('root.test.ExecutionTime', 'summary', quantiles=[0.5, 0.95],
                           count_calls=False, enable_timestamp=False)
        for i in range(9):
            metric.runtime(10.0)
        metric.runtime(15.0)

        self.assertTrue(MetricDataCollection(MetricData('ExecutionTime', 10.0, {'quantile': 0.5}),
                                             MetricData('ExecutionTime', 12.749999999999995,
                                                        {'quantile': 0.95})).strict_eq(
            metric.to_metric_data()))

    def test_labels(self):
        metric = getMetric('root.test.FloatValue', 'float', enable_timestamp=False)
        metric.runtime(2, label='value')
        metric.runtime(3, label='key')

        self.assertTrue(
            MetricDataCollection(MetricData('root.test.FloatValue', 2, {'label': 'value'}),
                                 MetricData('root.test.FloatValue', 3, {'label': 'key'})).strict_eq(
                metric.to_metric_data()))

    def test_base_metric(self):
        metric1 = getMetric('root.test.FloatValue', 'float', enable_timestamp=False)
        metric1.runtime(2.0)

        metric2 = getMetric('root.test.IntValue', 'int', enable_timestamp=False)
        metric2.runtime(3)

        root_metric = getMetric('', enable_timestamp=False)

        self.assertTrue(root_metric.to_metric_data().strict_eq(
            MetricData('root.test.FloatValue', 2),
            MetricData('root.test.IntValue', 3)))

    def test_wait(self):
        metric = getMetric('test', 'summary', quantiles=[0.5])

        @metric.measure()
        def wait(throw=False):
            time.sleep(1)
            if throw:
                raise ValueError('sample exception')

        wait()
        self.assertFalse(inspect.isgeneratorfunction(wait))
        self.assertRaises(ValueError, lambda: wait(throw=True))
        self.assertGreaterEqual(next(iter(metric.to_metric_data().values)).value, 1)

    def test_counter(self):
        counter = getMetric('counter', 'counter', enable_timestamp=False)
        counter.runtime(1, service='user')
        counter.runtime(2, service='session')
        counter.runtime(1, service='user')
        self.assertTrue(MetricDataCollection(MetricData('counter', 2, {'service': 'user'}),
                                             MetricData('counter', 2, {'service': 'session'}),
                                             MetricData('counter.sum', 4)).strict_eq(
            counter.to_metric_data()))

    def test_counter_count_calls(self):
        counter = getMetric('counter', 'counter', enable_timestamp=False, count_calls=True)
        counter.runtime(1, service='user')
        counter.runtime(2, service='session')
        counter.runtime(1, service='user')
        self.assertTrue(MetricDataCollection(MetricData('counter', 2, {'service': 'user'}),
                                             MetricData('counter.count', 2, {'service': 'user'}),
                                             MetricData('counter', 2, {'service': 'session'}),
                                             MetricData('counter.count', 1, {'service': 'session'}),
                                             MetricData('counter.sum', 4),
                                             MetricData('counter.count', 3)).strict_eq(
            counter.to_metric_data()))

    def test_base_metric(self):
        root_metric = getMetric()
        root_metric.level = MetricLevel.DEBUG
        metric2 = getMetric('root.test.FloatValue', 'float', MetricLevel.DEBUG, enable_timestamp=False)
        metric2.runtime(2.0)
        metric2.debug(1.0)

        metric3 = getMetric('root.test.IntValue', 'int', MetricLevel.RUNTIME, enable_timestamp=False)
        self.assertEqual(metric3.get_fully_qualified_name(), 'root.test.IntValue')
        metric3.runtime(3)
        metric3.debug(2)

        root_metric = getMetric('')

        self.assertTrue(
            MetricDataCollection(
                MetricData('root.test.FloatValue', 1.0),
                MetricData('root.test.IntValue', 3)).strict_eq(root_metric.to_metric_data()))

    def testInheritance(self):
        metric = getMetric('root.test.FloatValue', 'float', MetricLevel.INHERIT,
                           enable_timestamp=False)
        metric.runtime(2.0)
        metric_parent = getMetric('root.test', enable_timestamp=False)
        self.assertEqual(metric_parent.get_fully_qualified_name(), 'root.test')
        self.assertTrue(getMetric().to_metric_data().strict_eq(
            MetricDataCollection(MetricData('root.test.FloatValue', 2.0))))

        metric_parent.level = MetricLevel.RUNTIME
        metric.debug(3.0)

        self.assertTrue(getMetric('', enable_timestamp=False).to_metric_data().strict_eq(
            MetricDataCollection(MetricData('root.test.FloatValue', 2.0))))

    def test_labels(self):
        metric = getMetric('root.IntValue', 'int', labels={'k': 2}, enable_timestamp=False)
        metric.runtime(3)
        self.assertTrue(MetricDataCollection(MetricData('IntValue', 3, {'k': 2})).strict_eq(
            metric.to_metric_data()))

    def test_cps(self):
        metric = getMetric('root.CPSValue', 'cps', time_unit_vectors=[1, 2], enable_timestamp=False)
        self.assertEqual(metric.name, 'CPSValue')
        metric.runtime()
        self.assertEqual([1, 2], metric.time_unit_vectors)
        self.assertTrue(MetricDataCollection(MetricData('CPSValue', 1, {'period': 1}),
                                             MetricData('CPSValue', 1, {'period': 2})).strict_eq(
            metric.to_metric_data()))
        metric.runtime()
        self.assertTrue(MetricDataCollection(MetricData('CPSValue', 2, {'period': 1}),
                                             MetricData('CPSValue', 2, {'period': 2})).strict_eq(
            metric.to_metric_data()))
        time.sleep(1.2)
        self.assertTrue(MetricDataCollection(MetricData('CPSValue', 0, {'period': 1}),
                                             MetricData('CPSValue', 2, {'period': 2})).strict_eq(
            metric.to_metric_data()))

    def test_cps_labels(self):
        metric = getMetric('root.CPSValue', 'cps', time_unit_vectors=[1], enable_timestamp=False)
        metric.runtime(key='value')
        self.assertTrue(
            MetricDataCollection(MetricData('CPSValue', 1, {'period': 1, 'key': 'value'}),
                                 MetricData('CPSValue.total', 1, {'period': 1})).strict_eq(
                metric.to_metric_data()))



from .classes import GenerationPolicy, StackFrame, StoredVariableValue
from .trace_back import Traceback, frame_from_traceback

__all__ = ['GenerationPolicy', 'StoredVariableValue', 'StackFrame', 'Traceback',
           'frame_from_traceback']



from .dictionaries import DictObject, apply_dict_object, DictionaryView, TwoWayDictionary, \
    DirtyDict, KeyAwareDefaultDict, ExpiringEntryDict, SelfCleaningDefaultDict, \
    CacheDict, ExclusiveWritebackCache, CountingDict, LRUCacheDict, DefaultDict
from .hashable_objects import HashableWrapper
from .heaps import Heap, SetHeap, TimeBasedHeap, TimeBasedSetHeap
from .immutable import Immutable, frozendict, NotEqualToAnything, NOT_EQUAL_TO_ANYTHING
from .mixins import OmniHashableMixin, ReprableMixin, StrEqHashableMixin, ComparableIntEnum, \
    HashableIntEnum, ComparableAndHashableBy, ComparableAndHashableByInt, ComparableEnum, \
    HashableMixin, ComparableAndHashableByStr, DictionaryEQAble
from .proxy import Proxy
from .queues import Subqueue
from .ranking import Ranking
from .singleton import Singleton, SingletonWithRegardsTo, get_instances_for_singleton, \
    delete_singleton_for
from .sorted_list import SortedList, SliceableDeque
from .sparse_matrix import SparseMatrix
from .typednamedtuple import typednamedtuple
from .lru import LRU
from .syncable_droppable import DBStorage, SyncableDroppable
from .tuples import Vector
from .push_iterable import PushIterable

__all__ = [
    'PushIterable',
    'Vector',
    'DBStorage', 'SyncableDroppable',
    'LRU',
    'LRUCacheDict',
    'NotEqualToAnything', 'NOT_EQUAL_TO_ANYTHING',
    'HashableMixin',
    'CountingDict', 'DictionaryEQAble',
    'get_instances_for_singleton', 'delete_singleton_for',
    'Subqueue',
    'ExclusiveWritebackCache',
    'DefaultDict',
    'ComparableEnum',
    'CacheDict',
    'KeyAwareDefaultDict',
    'Proxy',
    'ReprableMixin',
    'ComparableAndHashableByInt',
    'StrEqHashableMixin',
    'ComparableIntEnum',
    'ComparableAndHashableBy',
    'HashableIntEnum',
    'ComparableAndHashableByStr',
    'DirtyDict',
    'SortedList',
    'SelfCleaningDefaultDict',
    'SliceableDeque',
    'Ranking',
    'ExpiringEntryDict',
    'TwoWayDictionary',
    'typednamedtuple',
    'HashableWrapper',
    'DictionaryView',
    'frozendict',
    'SparseMatrix',
    'OmniHashableMixin',
    'Singleton',
    'SingletonWithRegardsTo',
    'DictObject',
    'apply_dict_object',
    'Immutable',
    'SetHeap', 'TimeBasedHeap', 'TimeBasedSetHeap', 'Heap'
]



from .cache_dict import MetrifiedCacheDict, MetrifiedLRUCacheDict, MetrifiedExclusiveWritebackCache
from .threadpool import MetrifiedThreadPoolExecutor

__all__ = ['MetrifiedCacheDict', 'MetrifiedThreadPoolExecutor', 'MetrifiedLRUCacheDict',
           'MetrifiedExclusiveWritebackCache']



import typing as tp

from satella.json import JSONAble
import enum


def jsonify(data: tp.Any) -> tp.Optional[tp.Union[str, int, float, list, dict]]:
    """
    Convert any data to a value that's serializable via JSON.

    Objects that are JSONAble will have their to_json() method called.

    Note that enums will be converted to their value.

    As a last resort, str() will be called on the object, and if that fails
    it will have repr() called on it

    :param data: data to convert to a jsonable
    :return: JSON-able data
    """
    if data is None:
        v = None
    elif isinstance(data, (int, float, str)):
        v = data
    elif isinstance(data, enum.Enum):
        v = data.value
    elif isinstance(data, JSONAble):
        v = jsonify(data.to_json())
    elif isinstance(data, tp.Mapping):
        new_mapping = {}
        for key in data:
            new_mapping[jsonify(key)] = jsonify(data[key])
        v = new_mapping
    elif isinstance(data, (tp.Iterable, tp.Iterator)):
        v = [jsonify(elem) for elem in data]
    else:
        try:
            v = str(data)
        except TypeError:
            v = repr(data)
    return v
